# -*- coding: utf-8 -*-
"""Data storages."""

import os
import datetime
from .utilities import detect_encoding, sanitize_string, is_list_of_integers, to_float
from .units import TimeUnit
from .datastructures import DataTimePoint, DataTimePointSeries, TimePointSeries, TimePoint, DataTimeSlot, DataTimeSlotSeries
from .time import dt_from_str, dt_from_s, s_from_dt, timezonize, now_dt
from .exceptions import NoDataException, FloatConversionError


# Setup logging
import logging
logger = logging.getLogger(__name__)

HARD_DEBUG = False

POSSIBLE_TIMESTAMP_COLUMNS = ['timestamp', 'epoch']
NO_DATA_PLACEHOLDERS = ['na', 'nan', 'null', 'nd', 'undef']


#======================
#  CSV File Storage
#======================

class CSVFileStorage(object):
    """A CSV file data storage."""
    
    def __init__(self, filename_with_path, encoding = 'auto', time_column = 'auto', time_format = 'auto',
                 date_column = None, date_format = None, data_columns = 'all', value_separator=',', line_separator='\n',
                 skip_errors=False, data_format='auto', item_type=None, tz='UTC'):
        """Ref to https://docs.python.org/3/library/datetime.html#strftime-and-strptime-format-codes
        for date_format and time_format"""

        # File & encoding
        self.filename_with_path = filename_with_path
        if encoding == 'auto':
            self.encoding = None
        else:
            self.encoding = encoding

        # Time
        self.time_column = time_column
        self.time_format = time_format
        self.date_column = date_column
        self.date_format = date_format 
        
        # Time zone
        self.tz = tz   
        
        # Data
        self.data_columns = data_columns
        
        # Value and line separators
        self.value_separator = value_separator
        self.line_separator = line_separator

        # Other
        self.skip_errors = skip_errors
        self.data_format = data_format
        
        # Set item type (that will be forced)
        self.item_type = item_type

        # Check
        if self.time_column is None and self.date_column is None and not self.data_columns:
            raise ValueError('No time column, date column or data columns provided, cannot get anyything fromt this CSV file')
        
        if self.time_column is not None and not isinstance(self.time_column, int) and not isinstance(self.time_column, str):
            raise ValueError('time_column argument must be either integer os string (got "{}")'.format(self.time_column.__class__.__name__))

        if self.date_column is not None and not isinstance(self.date_column, int) and not isinstance(self.date_column, str):
            raise ValueError('date_column argument must be either integer os string (got "{}")'.format(self.date_column.__class__.__name__))

        if self.data_columns != 'all' and not isinstance(self.data_columns, list):
            raise ValueError('data_columns argument must be a list (got "{}")'.format(self.data_columns.__class__.__name__))


    def get(self, limit=None, tz=None, item_type=None, sort=False):

        # Line counter
        line_number=0
        
        # Init column indexes and labels
        column_indexes = None
        column_labels = None
        
        time_column_index = None
        date_column_index = None
        
        data_column_indexes = None
        data_column_labels = None

        # Detect encoding if not set
        if not self.encoding:
            self.encoding = detect_encoding(self.filename_with_path, streaming=False)

        items = []

        # TODO: evaluate rU vs newline='\n'
        with open(self.filename_with_path, 'r', encoding=self.encoding) as csv_file:

            while True:
                
                # Increase line_counter and check for limit
                line_number+=1
                if limit and line_number > limit:
                    break
                
                # TODO: replace me using a custom line separator
                line = csv_file.readline()
                logger.debug('Processing line #%s: "%s"', line_number, sanitize_string(line,NO_DATA_PLACEHOLDERS))

                # Do we have to stop? Note: empty lines still have the "\n" char.
                if not line:
                    break
                
                # Is this line an empty line?
                #if not line.strip():
                #    continue 
                                
                # Is this line a comment? 
                comment_chars = ['#', ';'] 
                if line[0] in comment_chars:
                    logger.debug('Skipping line "%s" as marked as comment line', line)
                    continue

                # Is this line something else? 
                if not self.value_separator in line:
                    logger.debug('Skipping line "%s" as marked no value separator found', line)
                    continue

                # Get values from this line:
                line_items = line.split(self.value_separator)

                # Set column keys if not already done
                if not column_indexes:
                    
                    # Is this line carrying non-numerical or non timestamp data?
                    not_converted = 0
                    for value in line_items:
                        try:
                            float(sanitize_string(value,NO_DATA_PLACEHOLDERS))
                        except:
                            try:
                                dt_from_str(sanitize_string(value,NO_DATA_PLACEHOLDERS))
                            except:
                                not_converted += 1
          
                    # If it is, use it as labels (and continue with the next line)
                    if not_converted: # == len(line_items):
                        column_indexes = [i for i in range(len(line_items)) if sanitize_string(line_items[i],NO_DATA_PLACEHOLDERS)]
                        column_labels = [sanitize_string(label,NO_DATA_PLACEHOLDERS) for label in line_items if sanitize_string(label,NO_DATA_PLACEHOLDERS)]
                        logger.debug('Set column indexes = "%s"  and column labels = "%s"', column_indexes, column_labels)
                        continue
                    
                    # Otherwise, just use integer keys (and go on)
                    else:
                        column_indexes = [i for i in range(len(line_items)) if sanitize_string(line_items[i],NO_DATA_PLACEHOLDERS)]
                        logger.debug('Set column indexes = "%s"  and column labels = "%s"', column_indexes, column_labels)

                    # Note: above we only used labels and indexes that carry content after being sanitize_stringd.
                    # TODO: what if a file starts with a line missing last value and we do or not do set data columns from the outside?
                    #       .. external data columns should have precedenze, but a mapping is then required more than just a list.

                #====================
                #  Time part
                #====================

                # Set time column index if not already done
                if time_column_index is None:
                    if self.time_column=='auto':
                        # TODO: improve this auto-detect: try different column names and formats.
                        #       Maybe, fix position as the first element for the timestamp.
                        if column_labels:
                            for posssible_time_column_name in POSSIBLE_TIMESTAMP_COLUMNS:
                                if posssible_time_column_name in column_labels:
                                    time_column_index = column_labels.index(posssible_time_column_name)
                                    self.time_column = posssible_time_column_name
                                    break                                
                            if time_column_index is None:
                                #raise Exception('Cannot auto-detect timestamp column') 
                                time_column_index = 0
                        else:
                            time_column_index = 0
                        # TODO: Try to convert to timestamp here?
                    else:
                        if isinstance(self.time_column, int):
                            time_column_index = self.time_column
                        else:
                            if self.time_column in column_labels:
                                time_column_index = column_labels.index(self.time_column)
                            else:
                                raise Exception('Cannot find requested time column "{}" in labels (got "{}")'.format(self.time_column, column_labels))                             
                    logger.debug('Set time column index = "%s"', time_column_index)

                # Set date column index if required and not already done
                if self.date_column is not None and date_column_index is None:
                    if isinstance(self.date_column, int):
                        date_column_index = self.date_column
                    else:
                        if self.date_column in column_labels:
                            date_column_index = column_labels.index(self.date_column)
                        else:
                            raise Exception('Cannot find requested time column "{}" in labels (got "{}")'.format(self.time_column, column_labels))                             
                    logger.debug('Set date column index = "%s"', date_column_index)


                # Time part
                time_part=''
                if time_column_index is not None:
                    time_part = sanitize_string(line_items[time_column_index],NO_DATA_PLACEHOLDERS)


                # Date part
                date_part=''
                if date_column_index is not None:
                    date_part = sanitize_string(line_items[date_column_index],NO_DATA_PLACEHOLDERS)
                                   
                # Assemble timestamp
                if self.date_column is not None:
                    timestamp = date_part + '\t' + time_part
                else:
                    timestamp = time_part
                  
                # Convert timestamp to epoch seconds
                logger.debug('Will process timestamp "%s"', timestamp)
                try:
                    
                    if self.time_format == 'auto':
                        logger.debug('Auto-detecting timestamp format')

                        # Is this an epoch?
                        try:
                            t = float(timestamp)
                            logger.debug('Auto-detected timestamp format: epoch')
                            self.time_format = 'epoch'
                        except:
                            try:
                                # is this an iso8601?
                                t = s_from_dt(dt_from_str(timestamp))
                                logger.debug('Auto-detected timestamp format: iso8601')
                                self.time_format = 'iso8601'
                            except:
                                raise Exception('Cannot auto-detect timestamp format (got "{}")'.format(timestamp)) from None
                    
                    elif self.time_format == 'epoch':
                        t = float(timestamp)

                    elif self.time_format == 'iso8601':
                        t = s_from_dt(dt_from_str(timestamp))
                    
                    else:
                        if self.date_column is not None:
                            dt = datetime.datetime.strptime(timestamp, self.date_format + '\t' + self.time_format)
                            t = s_from_dt(dt)
                        else:
                            dt = datetime.datetime.strptime(timestamp, self.time_format)
                            t = s_from_dt(dt)
    
                        if not dt.tzinfo:
                            # TODO: we should not do this check here, but above.
                            # Also maybe is not necessary at all if s_from_dt accepts timezone-unaware datetimes.
                            import pytz
                            dt = pytz.timezone('UTC').localize(dt)
                            #t = s_from_dt(dt)
                except Exception as e:
                    if self.skip_errors:
                        logger.debug('Cannot parse timestamp "%s" (%s) in line #%s, skipping the line.', timestamp, e, line_number)
                        continue
                    else:
                        raise Exception('Cannot parse timestamp "{}" ({}) in line #%{}, aborting. Set "skip_errors=True" to drop them instead.'.format(timestamp, e, line_number)) from None

                #====================
                #  Data part
                #====================
                
                if data_column_indexes is None:
                    
                    # Do we have to select only some data columns?
                    if self.data_columns != 'all':
                        
                        if is_list_of_integers(self.data_columns):
                            # Case where we have been given a list of integers
                            data_column_indexes =  self.data_columns
                            self.data_format = list

                        else:
                            # Case where we have been given a list of labels
                            if not column_labels:
                                raise Exception('You asked for data column labels but there are no labels in this CSV file')
                            data_column_indexes = []
                            for data_column in self.data_columns:
                                if data_column not in column_labels:
                                    raise Exception('Cannot find data column "{}" in CSV columns "{}"'.format(data_column, column_labels))
                                # What is the position of this requested data column in the CSV columns?
                                data_column_indexes.append(column_labels.index(data_column))

                    else:
                        # Case where we have to automatically set data column indexes to include all data columns
                        
                        # Initialize the data_column_indexes equal to the column_indexes
                        data_column_indexes = column_indexes
                        
                        # Remove time and date indexes from the data_column_indexes
                        if time_column_index is not None:
                            data_column_indexes.remove(time_column_index)
                        
                        if date_column_index is not None:
                            data_column_indexes.remove(date_column_index)

                    # Ok, now based on the data_column_indexes fill the data_column_labels if we have column labels
                    if column_labels:
                        data_column_labels=[]
                        for data_column_index in data_column_indexes:
                            data_column_labels.append(column_labels[data_column_index])
                            
                    logger.debug('Set data_column_indexes="%s" and data_column_labels="%s"', data_column_indexes, data_column_labels)


                # Set data
                try:
                    if data_column_labels:
                        if len(data_column_labels) >1 and self.data_format == float:
                            raise Exception('Requested data format as float but got more than 1 value')
                        if self.data_format == float:
                            data = to_float(line_items[data_column_indexes[0]],NO_DATA_PLACEHOLDERS)
                        elif self.data_format == list:
                            data = [to_float(line_items[index],NO_DATA_PLACEHOLDERS) for index in data_column_indexes]
                        else:
                            data = {column_labels[index]: to_float(line_items[index],NO_DATA_PLACEHOLDERS,column_labels[index]) for index in data_column_indexes}
                    else:
                        # Default here is to set a float if there is only one value.
                        # TODO: are we sure we want this?
                        if len (data_column_indexes) >1 or self.data_format == list:
                            data = [to_float(line_items[index],NO_DATA_PLACEHOLDERS) for index in data_column_indexes]
                        else:
                            data = to_float(line_items[data_column_indexes[0]],NO_DATA_PLACEHOLDERS)
           
                # TODO: here we drop the entire line. Instead, should we use a "None" and allow "Nones" in the DataPoints data? 
                except FloatConversionError as e:
                    if self.skip_errors:
                        logger.debug('Cannot convert value "%s" in line #%s to float, skipping the line.', e, line_number)
                        continue
                    else:
                        raise Exception('Cannot convert value "{}" in line #{} to float, aborting. Set "skip_errors=True" to drop them instead.'.format(e, line_number)) from None

                except IndexError as e:
                    if self.skip_errors:
                        logger.debug('Cannot parse in line #%s as some values are missing, skipping the line.', line_number)
                        continue
                    else:
                        raise Exception('Cannot parse line #{} as some values are missing, aborting. Set "skip_errors=True" to drop them instead.'.format(line_number)) from None

                logger.debug('Set data to "%s"', data)

                # Append to the series
                items.append([t, data])


        # Were we able to read something?
        if not items:
            raise NoDataException('Cannot read any data!')
        
        # Support var
        autodetect_item_type = False

        # Set item type if forced
        if not item_type and self.item_type:
            item_type = self.item_type

        # Otherwise, auto-detect
        if not item_type or item_type=='slots':
    
            autodetect_item_type = True
    
            # Detect sampling interval to create right item type (points or slots)
            from .utilities import detect_sampling_interval
            
            sample_timepoints = [TimePoint(t=item[0]) for item in items[0:10]]
            sample_timeseries = TimePointSeries(*sample_timepoints) # Cut to first ten elements
            detected_sampling_interval = detect_sampling_interval(sample_timeseries)
            
            # Years
            if detected_sampling_interval in [86400*365, 886400*366]:
                detected_item_type = DataTimeSlot
                detected_unit = TimeUnit('1Y') 
                   
            # Months
            elif detected_sampling_interval in [86400*31, 86400*30, 86400*28]:
                detected_item_type = DataTimeSlot
                detected_unit = TimeUnit('1M')
            
            # Days
            elif detected_sampling_interval in [3600*24, 3600*23, 3600*25]:
                detected_item_type = DataTimeSlot
                detected_unit = TimeUnit('1D')
            
            # Weeks still to be implemented in the unit
            #elif detected_sampling_interval in [3600*24*7, (3600*24*7)-3600, (3600*24*7)+3600]:
            #    detected_item_type = DataTimeSlot
            #    detected_unit = TimeUnit('1D')
            
            # Else, use points with no unit if we were not using slots
            else:
                if item_type!='slots':
                    detected_item_type = DataTimePoint
                    detected_unit = None
                else:
                    # TODO: "detected_item_type" is not a nice name here, the
                    # code in the following should use item_type if forced..
                    detected_item_type = DataTimeSlot

                    # Can we auto-detect unit? TODO: can we standardize this? Check also in the entire codebase..
                    if detected_sampling_interval == 3600:
                        detected_unit = TimeUnit('1h')                    
                    elif detected_sampling_interval == 1800:
                        detected_unit = TimeUnit('30m')
                    elif detected_sampling_interval == 900:
                        detected_unit = TimeUnit('15m')
                    elif detected_sampling_interval == 600:
                        detected_unit = TimeUnit('10m')                                             
                    elif detected_sampling_interval == 300:
                        detected_unit = TimeUnit('5m') 
                    elif detected_sampling_interval == 60:
                        detected_unit = TimeUnit('1m') 
                    else:
                        detected_unit = TimeUnit('{}s'.format(detected_sampling_interval)) 
                        
        # Do we have to force a specific type?
        if not autodetect_item_type:
            if item_type == 'points':
                item_type = DataTimePoint
                unit = None 
            elif item_type == 'slots':
                item_type = DataTimeSlot
                unit = TimeUnit('{}s'.format(detected_sampling_interval))
            else:
                raise ValueError('Unknown value "{}" for type. Accepted types are "points" or "slots".'.format(self.item_type))
        else:
            # Log the type and unit we detected 
            if detected_item_type == DataTimeSlot:
                if item_type:
                    logger.info('Assuming {} time unit and creating Slots.'.format(detected_unit))                
                else:
                    logger.info('Assuming {} time unit and creating Slots. Use item_type=\'points\' if you want Points instead.'.format(detected_unit))
            #else:
            #    logger.info('Assuming {} sampling interval and creating {}.'.format(detected_sampling_interval, item_type.__class__.__name__))
            
            # and use it.
            item_type = detected_item_type
            unit = detected_unit

        # Set and timezonize the timezone. In this way the it will be just a pointer.
        if tz:
            tz = timezonize(tz)
        else:
            tz = timezonize(self.tz)

        
        # Create point or slot series
        if item_type == DataTimePoint:
            timeseries = DataTimePointSeries()
            if sort:
                from operator import itemgetter
                items = sorted(items, key=itemgetter(0))
            for item in items:
                try:
                    data_loss = None
                    indexes = None
                    if isinstance (item[1], dict):
                        # Handle indexes and special data loss case
                        indexes = {}
                        for key in item[1]:
                            if key.startswith('__'):
                                if key=='__data_loss':
                                    data_loss = item[1][key]
                                else:
                                    indexes[key] = item[1][key]
                                    
                        # Remove indexes and data loss from item data
                        for index in indexes:       
                            item[1].pop(index)
                        item[1].pop('__data_loss',None)   
                                
                    # Create DataTimePoint, set data loss and set indexes
                    data_time_point = DataTimePoint(t=item[0], data=item[1], data_loss=data_loss, tz=tz)
                    if indexes:
                        for index in indexes: 
                            # Set index. The [2:] removes the two trailing underscores
                            setattr(data_time_point, index[2:], indexes[index])
                            
                    # Append
                    timeseries.append(data_time_point)

                except Exception as e:
                    if self.skip_errors:
                        logger.error(e)
                    else:
                        raise e from None
        else:
            
            timeseries = DataTimeSlotSeries()
            
            for i, item in enumerate(items):
                try:
                    data_loss = 0
                    if isinstance (item[1], dict):
                        # Handle indexes and special data loss case
                        indexes = {}
                        for key in item[1]:
                            if key.startswith('__'):
                                if key=='__data_loss':
                                    data_loss = item[1][key]
                                else:
                                    indexes[key] = item[1][key]
                                    
                        # Remove indexes and data loss from item data
                        for index in indexes:       
                            item[1].pop(index)
                        item[1].pop('__data_loss',None)     
                            
                    # Create DataTimeSlot, set data loss and set indexes
                    data_time_slot = DataTimeSlot(t=item[0], unit=unit, data=item[1], data_loss=data_loss, tz=tz)
                    if indexes:
                        for index in indexes: 
                            # Set index. The [2:] removes the two trailing underscores
                            setattr(data_time_slot, index[2:], indexes[index])
                            
                    timeseries.append(data_time_slot)
                    
                except ValueError as e:
                    # The only ValueError that could (should) arise here is a "Not in succession" error.
                    missing_timestamps = []
                    prev_dt = dt_from_s(items[i-1][0], tz=tz)
                    while True:                        
                        dt = prev_dt + unit
                        # Note: the equal here is just to prevent endless loops, the check shoudl actally be just an equal     
                        if s_from_dt(dt) >= item[0]:
                            # We are arrived, append all the missing items and then the item we originally tried to and break
                            for j, missing_timestamp in enumerate(missing_timestamps):
                                # Set data by interpolation
                                interpolated_data = {data_key: (((items[i][1][data_key]-items[i-1][1][data_key])/(len(missing_timestamps)+1)) * (j+1)) + items[i-1][1][data_key]  for data_key in  items[-1][1] }
                                timeseries.append(DataTimeSlot(t=missing_timestamp, unit=unit, data=interpolated_data, data_loss=1, tz=tz))
                            timeseries.append(DataTimeSlot(t=item[0], unit=unit, data=item[1], data_loss=0, tz=tz))
                            break
                        
                        else:
                            missing_timestamps.append(s_from_dt(dt))
                        prev_dt = dt

        return timeseries


    def put(self, timeseries, overwrite=False):
        
        if os.path.isfile(self.filename_with_path) and not overwrite:
            raise Exception('File already exists. use overwrite=True to overwrite.')

        # Set indexes here once for all or thre will be a slowndowd afterwards
        indexes = timeseries.indexes

        with open(self.filename_with_path, 'w') as csv_file:
       
            # 0) Dump CSV-storage specific metadata & chck type     
            if isinstance(timeseries, DataTimePointSeries):
                csv_file.write('# Generated by Timeseria CSVFileStorage at {}\n'.format(now_dt()))
                csv_file.write('# {}\n'.format(timeseries))
                csv_file.write('# Extra parameters: {{"type": "points", "tz": "{}", "resolution": "{}"}}\n'.format(timeseries.tz, timeseries.resolution))
 
            elif isinstance(timeseries, DataTimeSlotSeries):                
                csv_file.write('# Generated by Timeseria CSVFileStorage at {}\n'.format(now_dt()))
                csv_file.write('# {}\n'.format(timeseries))
                csv_file.write('# Extra parameters: {{"type": "slots", "tz": "{}", "resolution": "{}"}}\n'.format(timeseries.tz, timeseries.resolution))

            else:
                raise TypeError('Can store only DataTimePointSeries or DataTimeSlotSeries')


            # 1) Dump headers
            data_keys_part = ','.join([str(key) for key in timeseries.data_keys()])
            indexes_part = ','.join(['__'+index for index in indexes])
            if indexes_part:
                csv_file.write('epoch,{},{}\n'.format(data_keys_part,indexes_part))
            else:
                csv_file.write('epoch,{}\n'.format(data_keys_part))
            

            # 2) Dump data (and indexes)
            for item in timeseries:
                data_part = ','.join([str(item.data[key]) for key in timeseries.data_keys()])
                indexes_part = ','.join([str(getattr(item, index)) for index in indexes])
                if indexes_part:
                    csv_file.write('{},{},{}\n'.format(item.t, data_part, indexes_part))
                else:
                    csv_file.write('{},{}\n'.format(item.t, data_part))
                    










# -*- coding: utf-8 -*-
"""Operations on the series, returning both scalar and other series."""

from copy import copy, deepcopy
from .time import s_from_dt
from .datastructures import Series, Slot, Point, PointSeries
from .utilities import is_close
from .units import TimeUnit, Unit

# Setup logging
import logging
logger = logging.getLogger(__name__)


#=======================
#  Base Operation class
#=======================

class Operation():
    """Generic operation class (callable object)."""

    @classmethod
    def __str__(cls):
        return '{} operation'.format(cls.__name__)
    
    @classmethod
    def __repr__(cls):
        return '{}'.format(cls.__name__.lower())
    
    def __call__(self, *args, **kwargs):
        raise NotImplementedError('No operation logic implemented.')
    
    @property
    def __name__(self):
        return self.__class__.__name__.lower()


class ScalarOperation(Operation):
    """An operation operating on a series and returning a scalar (callable object)."""
    pass


class SeriesOperation(Operation):
    """An operation operating on a series and returning a series (callable object)."""
    pass


#=======================
#  Scalar Operations
#=======================

class Max(ScalarOperation):
    """Maximum operation (callable object). Comes also pre-instantiated as the ``max()``
    function in the same module (accessible as ``timeseria.operations.max``)."""
    
    def __init__(self):
        self.built_in_max = max
    
    def __call__(self, arg, key=None, **kwargs):
        if not isinstance(arg, Series):
            return self.built_in_max(arg)
        else:
            series = arg
            mins = {key: None for key in series.data_keys()}
            for item in series:
                for data_key in mins:
                    if mins[data_key] is None:
                        mins[data_key] = item.data[data_key]
                    else:
                        if item.data[data_key] > mins[data_key]:
                            mins[data_key] = item.data[data_key]
            
            if key is not None:
                return mins[key]
            if len(mins) == 1:
                return mins[series.data_keys()[0]]
            else:
                return mins


class Min(ScalarOperation):
    """Minimum operation (callable object). Comes also pre-instantiated as the ``min()``
    function in the same module (accessible as ``timeseria.operations.min``)."""
    
    def __init__(self):
        self.built_in_min = min
    
    def __call__(self, arg, key=None, **kwargs):
        if not isinstance(arg, Series):
            return self.built_in_min(arg)
        else:
            series = arg
            mins = {key: None for key in series.data_keys()}
            for item in series:
                for data_key in mins:
                    if mins[data_key] is None:
                        mins[data_key] = item.data[data_key]
                    else:
                        if item.data[data_key] < mins[data_key]:
                            mins[data_key] = item.data[data_key]
            
            if key is not None:
                return mins[key]
            if len(mins) == 1:
                return mins[series.data_keys()[0]]
            else:
                return mins


class Avg(ScalarOperation):
    """Average operation (callable object)."""
    
    def __call__(self, arg, key=None, **kwargs):
        if not isinstance(arg, Series):
            raise NotImplementedError('Avg implemented only on series objects')
        else:
            series = arg
            sums = {key: None for key in series.data_keys()}
            for item in series:
                for data_key in sums:
                    if sums[data_key] is None:
                        sums[data_key] = item.data[data_key]
                    else:
                        sums[data_key] += item.data[data_key]
            avgs={}
            for data_key in sums:
                avgs[data_key] = sums[data_key] / len(series)
            if key is not None:
                return avgs[key]
            if len(avgs) == 1:
                return avgs[series.data_keys()[0]]
            else:
                return avgs


class WAvg(ScalarOperation):
    """Weighted average operation (callable object)."""

    def __call__(self, series, prev_point=None, next_point=None):
        raise NotImplementedError('Weighted Average is not yet implemented')
        #keys = None
        #avgs = {}
        #count=0
        #for data_time_point in series:
        #    count+=1              
        #    if not keys:
        #        keys=data_time_point.data.keys()
        #    for key in keys:
        #        if key not in avgs:
        #            avgs[key] = 0
        #        avgs[key] += data_time_point.data[key]
        #for key in keys:
        #    new_avgs = {key:avgs[key]/count for key in keys}
        #return new_avgs


class Sum(ScalarOperation):
    """Sum operation (callable object)."""

    def __init__(self):
        self.built_in_sum = sum
    
    def __call__(self, arg, key=None, **kwargs):
        if not isinstance(arg, Series):
            return self.built_in_sum(arg)
        else:
            series = arg
            sums = {key: 0 for key in series.data_keys()}
            for item in series:
                for data_key in sums:
                    sums[data_key] += item.data[data_key]
            if key is not None:
                return sums[key]
            if len(sums) == 1:
                return sums[series.data_keys()[0]]
            else:
                return sums


#=======================
#  Series operations
#=======================

class Derivative(SeriesOperation):
    """Derivative operation (callable object). Must operate on a fixed resolution time series"""
    
    def __call__(self, timeseries, inplace=False, normalize=True, diffs=False):
        
        if normalize or not diffs:
            if isinstance(timeseries, PointSeries) and timeseries.resolution == 'variable':
                raise ValueError('Variable resolutions are not supported in this mode. Resample or slot the time series first.')
            postfix='derivative'
        else:
            postfix='diff'

        if not inplace:
            der_timeseries = timeseries.__class__()

        data_keys = timeseries.data_keys()
        for i, item in enumerate(timeseries):

            if not inplace:
                data = {}
            
            for key in data_keys:
                
                if diffs:
                    # Just compute the diffs
                    if i == 0:
                        diff = 0
                    else:
                        diff = timeseries[i].data[key] - timeseries[i-1].data[key]
                else:
                    # Compute the derivative
                    if i == 0:
                        # Right increment for the first item
                        diff = timeseries[i+1].data[key] - timeseries[i].data[key]
                    elif i == len(timeseries)-1:
                        # Left increment for the last item. Divide by two if not normlaizing by time resolution
                        diff = timeseries[i].data[key] - timeseries[i-1].data[key]
                    else:
                        # Both left and right increment for the items in the middle, averaged.
                        diff =  ((timeseries[i+1].data[key] - timeseries[i].data[key]) + (timeseries[i].data[key] - timeseries[i-1].data[key]))
                        diff = diff/2

                # Normalize the increment to get the actual derivative
                if normalize:
                    if isinstance(timeseries.resolution, TimeUnit):
                        diff = diff / timeseries.resolution.duration_s(item.dt)               
                    elif isinstance(timeseries.resolution, Unit):
                        diff = diff / timeseries.resolution.value
                    else:
                        diff = diff / timeseries.resolution

                # Add data
                if not inplace:
                    data['{}_{}'.format(key, postfix)] = diff
                else:
                    item.data['{}_{}'.format(key, postfix)] = diff
            
            # Create the item
            if not inplace:

                if isinstance(timeseries[0], Point):
                    der_timeseries.append(timeseries[0].__class__(t = item.t,
                                                                  tz = item.tz,
                                                                  data = data))         
                elif isinstance(timeseries[0], Slot):
                    der_timeseries.append(timeseries[0].__class__(start = item.start,
                                                                  unit = item.unit,
                                                                  data = data))                
                else:
                    raise NotImplementedError('Working on series other than slots or points not yet implemented')

        if not inplace:
            return der_timeseries


class Integral(SeriesOperation):
    """Integral operation (callable object). Must operate on a fixed resolution time series."""
    
    def __call__(self, timeseries, inplace=False, normalize=True, c=0):
        
        if normalize:
            if isinstance(timeseries, PointSeries) and timeseries.resolution == 'variable':
                raise ValueError('Variable resolutions are not supported. Resample or slot the time series first.')
            postfix='integral'
        else:
            postfix='csum'

        if not inplace:
            der_timeseries = timeseries.__class__()

        last_values={}

        data_keys = timeseries.data_keys()
        for i, item in enumerate(timeseries):

            if not inplace:
                data = {}
            
            for key in data_keys:
                
                # Get the value
                if not normalize:
                    value = timeseries[i].data[key]
                else:
                    # Compute the integral
                    if i == 0:
                        value=0
                        prev_right_component = timeseries[i].data[key]
                    else:
                        left_component = (timeseries[i-1].data[key]*2) - prev_right_component
                        # The "value" below is actually an increment, see below at  value = last_values[key] + value 
                        value = left_component 
                        prev_right_component = left_component

                # Normalize the cumulative to get the actual integral
                if normalize:
                    if isinstance(timeseries.resolution, TimeUnit):
                        value = value * timeseries.resolution.duration_s(item.dt)               
                    elif isinstance(timeseries.resolution, Unit):
                        value = value * timeseries.resolution.value
                    else:
                        value = value * timeseries.resolution

                # Sum to previous or apply offset if any
                try:
                    value = last_values[key] + value
                except:
                    try:
                        # Do we have offsets as a dict?
                        value = value + c[key]
                    except:
                        value = value + c               

                # Add data
                if not inplace:
                    data['{}_{}'.format(key, postfix)] = value
                else:
                    item.data['{}_{}'.format(key, postfix)] = value
                
                last_values[key] = value

           
            # Create the item
            if not inplace:

                if isinstance(timeseries[0], Point):
                    der_timeseries.append(timeseries[0].__class__(t = item.t,
                                                                  tz = item.tz,
                                                                  data = data))         
                elif isinstance(timeseries[0], Slot):
                    der_timeseries.append(timeseries[0].__class__(start = item.start,
                                                                  unit = item.unit,
                                                                  data = data))                
                else:
                    raise NotImplementedError('Working on series other than slots or points not yet implemented')

        if not inplace:
            return der_timeseries


class Diff(Derivative):
    """Incremental differences operation (callable object)."""
    def __call__(self, timeseries, inplace=False):
        return super(Diff, self).__call__(timeseries, inplace=inplace, normalize=False, diffs=True)


class CSum(Integral):
    """Cumulative sum operation (callable object)."""
    def __call__(self, timeseries, inplace=False, offset=0):
        return super(CSum, self).__call__(timeseries, inplace=inplace, normalize=False, c=offset)


class MAvg(SeriesOperation):
    """Moving average operation (callable object)."""

    def __call__(self, timeseries, window, inplace=False):

        if not window or window <1:
            raise ValueError('A integer window >0 is required (got window="{}"'.frmat(window))

        if inplace:
            raise Exception('In-place operation not supported yet (padding support required)')
        
        postfix='mavg_{}'.format(window)
        
        mavg_timeseries = timeseries.__class__()

        data_keys = timeseries.data_keys()
        for i, item in enumerate(timeseries):

            if i < window-1:
                continue

            data = {}

            for key in data_keys:
 
                # Compute the moving average
                value_sum = 0
                for j in range(i-window+1,i+1):
                    value_sum += timeseries[j].data[key]
                data['{}_{}'.format(key, postfix)] = value_sum/window

            # Create the item
            if isinstance(timeseries[0], Point):
                mavg_timeseries.append(timeseries[0].__class__(t = item.t,
                                                               tz = item.tz,
                                                               data = data))         
            elif isinstance(timeseries[0], Slot):
                mavg_timeseries.append(timeseries[0].__class__(start = item.start,
                                                               unit = item.unit,
                                                               data = data))                
            else:
                raise NotImplementedError('Working on series other than slots or points not yet implemented')

        return mavg_timeseries


class Select(SeriesOperation):
    """Select operation (callable object). Selects items of the series given SQL-like queries."""
    
    def __call__(self, series, query, *args, **kwargs):

        if (' and ' or ' AND ' or ' or ' or ' OR ') in query:
            raise NotImplementedError('Multiple conditions not yet supported')
        
        if '=' not in query:
            raise NotImplementedError('Clauses other than the equality are not yet supported')
        
        # Get key
        key = query.split('=')[0].strip()
        if key.startswith('"'):
            key=key[1:]
        if key.endswith('"'):
            key=key[:-1]
        
        # Get value
        value = float(query.split('=')[1])

        # Perform the select
        selected = None
        for item in series:
            if item.data[key] == value:
                try:
                    selected.append(item)
                except:
                    selected = [item]
        return selected 


class Filter(SeriesOperation):
    """Filter a series (callable object)."""

    def __call__(self, series, data_key=None, from_t=None, to_t=None, from_dt=None, to_dt=None):
        if from_dt:
            if from_t is not None:
                raise Exception('Cannot set both from_t and from_dt')
            from_t = s_from_dt(from_dt)
        if to_dt:
            if to_t is not None:
                raise Exception('Cannot set both to_t and to_dt')
            to_t = s_from_dt(to_dt)
            
        if from_t is not None and to_t is not None:
            if from_t >= to_t:
                raise Exception('Got from >= to')
        
        # Instantiate the filtered series
        filtered_series = series.__class__()
        
        # Filter based on time if from/to are set
        for item in series:
            if from_t is not None and to_t is not None:
                if item.t >= from_t and item.t <to_t:
                    filtered_series.append(item)
            else:
                if from_t is not None:
                    if item.t >= from_t:
                        filtered_series.append(item) 
                if to_t is not None:
                    if item.t < to_t:
                        filtered_series.append(item)
                if from_t is None and to_t is None:
                    # Append everything 
                    filtered_series.append(item)
        
        # Filter based on data key if set
        if data_key:
            for i, item in enumerate(filtered_series):
                filtered_series[i] = copy(item)
                filtered_series[i]._data = {data_key: item.data[data_key]}
        
        # Re-set reference data as well
        try:
            filtered_series.item_data_reference
        except AttributeError:
            pass
        else:
            filtered_series.item_data_reference = filtered_series[0].data

        return filtered_series 


class Merge(SeriesOperation):
    """Merge two or more series (callable object)."""
    
    def __call__(self, *timeseriess):
        
        resolution = None
        for i, arg in enumerate(timeseriess):
            #if not isinstance(arg, DataTimeSlotSeries):
            #    raise TypeError('Argument #{} is not of type DataTimeSlotSeries, got "{}"'.format(i, arg.__class__.__name__))
            if resolution is None:
                resolution = arg.resolution
            else:
                if arg.resolution != resolution:
                    abort = True
                    try:
                        # Handle floating point precision issues 
                        if is_close(arg.resolution, resolution):
                            abort = False
                    except (ValueError,TypeError):
                        pass
                    if abort:
                        raise ValueError('DataTimeSlotSeries have different units, cannot merge')
        
        length = len(timeseriess[0])
        n_timeseriess = len(timeseriess)
        result_timeseries = timeseriess[0].__class__(unit=resolution)
        
        for i in range(length):
            data = None
            coverage = None
            valid_coverages = 0
            for j in range(n_timeseriess):
                #logger.critical('i={}, j={}'.format(i, j))
                
                # Data
                if data is None:
                    data = deepcopy(timeseriess[j][i].data)
                else:
                    data.update(timeseriess[j][i].data)
                
                # Coverage
                if coverage is None:
                    coverage = coverage
                else:
                    valid_coverages += 1
                    coverage += timeseriess[j][i].coverage
            
            # Finalize coverage if there were valid   
            if valid_coverages:
                coverage = coverage / valid_coverages
            else:
                coverage = None

            if isinstance(timeseriess[0][0], Point):
                result_timeseries.append(timeseriess[0][0].__class__(t = timeseriess[j][i].t,
                                                                     tz = timeseriess[j][i].tz,
                                                                     data  = data))         
            elif isinstance(timeseriess[0][0], Slot):
                result_timeseries.append(timeseriess[0][0].__class__(start = timeseriess[j][i].start,
                                                                     unit   = timeseriess[j][i].unit,
                                                                     data  = data))                
            else:
                raise NotImplementedError('Working on series other than slots or points not yet implemented')

        return result_timeseries



#========================
# Instantiate operations
#========================

# Scalar operations
min = Min()
max = Max()
avg = Avg()
wavg = WAvg()
sum = Sum()

# Series operations
derivative = Derivative()
integral = Integral()

diff = Diff()
csum = CSum()
mavg = MAvg()

merge = Merge()
filter = Filter()
select = Select()




import unittest
from ..time import dt
from ..units import Unit, TimeUnit

class TestUnits(unittest.TestCase):

    def setUp(self):       
        pass

    def test_Unit(self):
        
        with self.assertRaises(ValueError):
            _ = TimeUnit('hi')        
        
        unit = Unit(60)
        self.assertEqual(unit.value, 60)

    def test_Unit_math(self):
        
        # Sum & subtract
        unit1 = Unit(60)
        self.assertEqual(unit1+5, 65)
        self.assertEqual(5+unit1, 65)
        self.assertEqual(unit1-5, 55)
        self.assertEqual(5-unit1, -55)        

        # Sum & subtract Units
        unit2 = Unit(67)
        self.assertEqual(unit1+unit2, 127)
        self.assertEqual(unit1-unit2, -7)
        self.assertEqual(unit2-unit1, 7)
        
        # Divide & multiply
        self.assertEqual(unit1/2, 30)
        self.assertEqual(120/unit1,2)
        self.assertEqual(unit1*2,120)
        self.assertEqual(2*unit1,120)


    def test_TimeUnit(self):

        with self.assertRaises(ValueError):
            _ = TimeUnit('15m', '20s')

        # Not valid 'q' type
        with self.assertRaises(ValueError):
            _ = TimeUnit('15q')

        # Numerical init
        time_unit_1 = TimeUnit(60)
        self.assertEqual(str(time_unit_1), '60s')

        # String init
        time_unit_1 = TimeUnit('15m')
        self.assertEqual(str(time_unit_1), '15m')

        time_unit_2 = TimeUnit('15m_30s_3u')
        self.assertEqual(str(time_unit_2), '15m_30s_3u')
        
        time_unit_3 = TimeUnit(days=1)
        
        # Value (duration in seconds)
        self.assertEqual(time_unit_1.value, 900)

        with self.assertRaises(TypeError):
            time_unit_3.value # Not defined for calendar time units
        
        #=====================
        #  Sum
        #=====================
        
        # Sum with other TimeUnit objects
        self.assertEqual(str(time_unit_1+time_unit_2+time_unit_3), '1D_30m_30s_3u')

        # Sum with datetime (also on DST change)
        time_unit = TimeUnit('1h')
        date_time1 = dt(2015,10,25,0,15,0, tzinfo='Europe/Rome')
        date_time2 = date_time1 + time_unit
        date_time3 = date_time2 + time_unit
        date_time4 = date_time3 + time_unit
        date_time5 = date_time4 + time_unit

        self.assertEqual(str(date_time1), '2015-10-25 00:15:00+02:00')
        self.assertEqual(str(date_time2), '2015-10-25 01:15:00+02:00')
        self.assertEqual(str(date_time3), '2015-10-25 02:15:00+02:00')
        self.assertEqual(str(date_time4), '2015-10-25 02:15:00+01:00')
        self.assertEqual(str(date_time5), '2015-10-25 03:15:00+01:00')

        # Sum with a numerical value
        time_unit = TimeUnit('1h')
        epoch1 = 3600
        self.assertEqual(epoch1 + time_unit, 7200)


        #=====================
        #  Subtraction
        #=====================

        # Subtract to other TimeUnit object
        with self.assertRaises(NotImplementedError):
            time_unit_1 - time_unit_2

        # Subtract to a datetime object
        with self.assertRaises(NotImplementedError):
            time_unit_1 - date_time1
            
        # In general, subtracting to anything is not implemented
        with self.assertRaises(NotImplementedError):
            time_unit_1 - 'hello'
        
        # Subtract from a datetime (also on DST change)
        time_unit = TimeUnit('1h')
        date_time1 = dt(2015,10,25,3,15,0, tzinfo='Europe/Rome')
        date_time2 = date_time1 - time_unit
        date_time3 = date_time2 - time_unit
        date_time4 = date_time3 - time_unit
        date_time5 = date_time4 - time_unit

        self.assertEqual(str(date_time1), '2015-10-25 03:15:00+01:00')
        self.assertEqual(str(date_time2), '2015-10-25 02:15:00+01:00')
        self.assertEqual(str(date_time3), '2015-10-25 02:15:00+02:00')
        self.assertEqual(str(date_time4), '2015-10-25 01:15:00+02:00')        
        self.assertEqual(str(date_time5), '2015-10-25 00:15:00+02:00')

        # Subtract from a numerical value
        time_unit = TimeUnit('1h')
        epoch1 = 7200
        self.assertEqual(epoch1 - time_unit, 3600)


        #=====================
        #  Types
        #=====================

        # Test type
        self.assertEqual(TimeUnit('15m').type, TimeUnit._PHYSICAL)
        self.assertEqual(TimeUnit('1h').type, TimeUnit._PHYSICAL)
        self.assertEqual(TimeUnit('1D').type, TimeUnit._CALENDAR)
        self.assertEqual(TimeUnit('1M').type, TimeUnit._CALENDAR)
        
        # Test sum with TimePoint
        time_unit = TimeUnit('1h')
        from ..datastructures import TimePoint
        time_point = TimePoint(60)
        self.assertEqual((time_point+time_unit).t, 3660)

        # Test unit value
        with self.assertRaises(TypeError):
            TimeUnit('1D').value
        with self.assertRaises(TypeError):
            TimeUnit('2Y').value 
        self.assertEqual(TimeUnit('1m').value, 60)
        self.assertEqual(TimeUnit('15m').value, 900)
        self.assertEqual(TimeUnit('1h').value, 3600)

        # Test equal
        self.assertEqual(time_unit_1, 900)

        # Test duration with composite point seconds init
        self.assertEqual(TimeUnit(minutes=1, seconds=3).value, 63)
        
        # Test floating point seconds init
        self.assertEqual(TimeUnit('1.2345s').value, 1.2345)
        self.assertEqual(TimeUnit('1.234s').value, 1.234)
        self.assertEqual(TimeUnit('1.02s').value, 1.02)
        self.assertEqual(TimeUnit('1.000005s').value, 1.000005)
        self.assertEqual(TimeUnit('67.000005s').seconds, 67)
        self.assertEqual(TimeUnit('67.000005s').microseconds, 5)
  
        # Too much precision (below microseconds), gets cut
        time_unit = TimeUnit('1.0000005s')
        self.assertEqual(str(time_unit),'1s')
        time_unit = TimeUnit('1.0000065s')
        self.assertEqual(str(time_unit),'1s_6u')
        

    def test_TimeUnit_duration(self):

        date_time1 = dt(2015,10,24,0,15,0, tzinfo='Europe/Rome')
        date_time2 = dt(2015,10,25,0,15,0, tzinfo='Europe/Rome')
        date_time3 = dt(2015,10,26,0,15,0, tzinfo='Europe/Rome')
 
        # Day unit
        time_unit = TimeUnit('1D')
        self.assertEqual(time_unit.duration_s(date_time1), 86400) # No DST, standard day
        self.assertEqual(time_unit.duration_s(date_time2), 90000) # DST, change
                 
        # Week unit
        time_unit = TimeUnit('1W')
        self.assertEqual(time_unit.duration_s(date_time1), (86400*7)+3600)
        self.assertEqual(time_unit.duration_s(date_time3), (86400*7))
         
        # Month Unit
        time_unit = TimeUnit('1M')
        self.assertEqual(time_unit.duration_s(date_time1), ((86400*31)+3600)) # October has 31 days, but here we have a DST change in the middle
        self.assertEqual(time_unit.duration_s(date_time3), (86400*31)) # October has 31 days

        # Year Unit
        time_unit = TimeUnit('1Y')
        self.assertEqual(time_unit.duration_s(dt(2014,10,24,0,15,0, tzinfo='Europe/Rome')), (86400*365)) # Standard year
        self.assertEqual(time_unit.duration_s(dt(2015,10,24,0,15,0, tzinfo='Europe/Rome')), (86400*366)) # Leap year
        

    def test_TimeUnit_shift_dt(self):

        date_time1 = dt(2015,10,24,0,15,0, tzinfo='Europe/Rome')
        date_time2 = dt(2015,10,25,0,15,0, tzinfo='Europe/Rome')
        date_time3 = dt(2015,10,26,0,15,0, tzinfo='Europe/Rome')

        # Day unit
        time_unit = TimeUnit('1D')
        self.assertEqual(time_unit.shift_dt(date_time1), dt(2015,10,25,0,15,0, tzinfo='Europe/Rome')) # No DST, standard day
        self.assertEqual(time_unit.shift_dt(date_time2), dt(2015,10,26,0,15,0, tzinfo='Europe/Rome')) # DST, change
                 
        # Week unit
        time_unit = TimeUnit('1W')
        self.assertEqual(time_unit.shift_dt(date_time1), dt(2015,10,31,0,15,0, tzinfo='Europe/Rome'))
        self.assertEqual(time_unit.shift_dt(date_time3), dt(2015,11,2,0,15,0, tzinfo='Europe/Rome'))
         
        # Month Unit
        time_unit = TimeUnit('1M')
        self.assertEqual(time_unit.shift_dt(date_time1), dt(2015,11,24,0,15,0, tzinfo='Europe/Rome'))
        self.assertEqual(time_unit.shift_dt(date_time2), dt(2015,11,25,0,15,0, tzinfo='Europe/Rome'))
        self.assertEqual(time_unit.shift_dt(date_time3), dt(2015,11,26,0,15,0, tzinfo='Europe/Rome'))
        
        # Test 12%12 must give 12 edge case
        self.assertEqual(time_unit.shift_dt(dt(2015,1,1,0,0,0, tzinfo='Europe/Rome')), dt(2015,2,1,0,0,0, tzinfo='Europe/Rome'))
        self.assertEqual(time_unit.shift_dt(dt(2015,11,1,0,0,0, tzinfo='Europe/Rome')), dt(2015,12,1,0,0,0, tzinfo='Europe/Rome'))

        # Year Unit
        time_unit = TimeUnit('1Y')
        self.assertEqual(time_unit.shift_dt(date_time1), dt(2016,10,24,0,15,0, tzinfo='Europe/Rome'))
        

    def test_TimeUnit_math(self):

        # Test that complex time_units are not handable
        time_unit = TimeUnit('1D_3h_5m')
        date_time = dt(2015,1,1,16,37,14, tzinfo='Europe/Rome')
        
        with self.assertRaises(ValueError):
            _ = time_unit.floor_dt(date_time)

        # Test in ceil/floor/round normal conditions (hours)
        time_unit = TimeUnit('1h')
        date_time = dt(2015,1,1,16,37,14, tzinfo='Europe/Rome')
        self.assertEqual(time_unit.floor_dt(date_time), dt(2015,1,1,16,0,0, tzinfo='Europe/Rome'))
        self.assertEqual(time_unit.ceil_dt(date_time), dt(2015,1,1,17,0,0, tzinfo='Europe/Rome'))

        # Test in ceil/floor/round normal conditions (minutes)
        time_unit = TimeUnit('15m')
        date_time = dt(2015,1,1,16,37,14, tzinfo='Europe/Rome')
        self.assertEqual(time_unit.floor_dt(date_time), dt(2015,1,1,16,30,0, tzinfo='Europe/Rome'))
        self.assertEqual(time_unit.ceil_dt(date_time), dt(2015,1,1,16,45,0, tzinfo='Europe/Rome'))

        # Test ceil/floor/round in normal conditions (seconds)
        time_unit = TimeUnit('30s')
        date_time = dt(2015,1,1,16,37,14, tzinfo='Europe/Rome') 
        self.assertEqual(time_unit.floor_dt(date_time), dt(2015,1,1,16,37,0, tzinfo='Europe/Rome'))
        self.assertEqual(time_unit.ceil_dt(date_time), dt(2015,1,1,16,37,30, tzinfo='Europe/Rome'))

        # Test ceil/floor/round across 1970-1-1 (minutes) 
        time_unit = TimeUnit('5m')
        date_time1 = dt(1969,12,31,23,57,29, tzinfo='UTC') # epoch = -3601
        date_time2 = dt(1969,12,31,23,59,59, tzinfo='UTC') # epoch = -3601       
        self.assertEqual(time_unit.floor_dt(date_time1), dt(1969,12,31,23,55,0, tzinfo='UTC'))
        self.assertEqual(time_unit.ceil_dt(date_time1), dt(1970,1,1,0,0, tzinfo='UTC'))
        self.assertEqual(time_unit.round_dt(date_time1), dt(1969,12,31,23,55,0, tzinfo='UTC'))
        self.assertEqual(time_unit.round_dt(date_time2), dt(1970,1,1,0,0, tzinfo='UTC'))

        # Test ceil/floor/round (3 hours-test)
        time_unit = TimeUnit('3h')
        date_time = dt(1969,12,31,23,0,1, tzinfo='Europe/Rome') # negative epoch    
        self.assertEqual(time_unit.floor_dt(date_time), dt(1969,12,31,23,0,0, tzinfo='Europe/Rome'))
        self.assertEqual(time_unit.ceil_dt(date_time), dt(1970,1,1,2,0, tzinfo='Europe/Rome'))

        # Test ceil/floor/round across 1970-1-1 (together with the 2 hours-test, TODO: decouple) 
        time_unit = TimeUnit('2h')
        date_time1 = dt(1969,12,31,22,59,59, tzinfo='Europe/Rome') # negative epoch
        date_time2 = dt(1969,12,31,23,0,1, tzinfo='Europe/Rome') # negative epoch  
        self.assertEqual(time_unit.floor_dt(date_time1), dt(1969,12,31,22,0,0, tzinfo='Europe/Rome'))
        self.assertEqual(time_unit.ceil_dt(date_time1), dt(1970,1,1,0,0, tzinfo='Europe/Rome'))
        self.assertEqual(time_unit.round_dt(date_time1), dt(1969,12,31,22,0, tzinfo='Europe/Rome'))
        self.assertEqual(time_unit.round_dt(date_time2), dt(1970,1,1,0,0, tzinfo='Europe/Rome'))

        # Test ceil/floor/round across DST change (hours)
        time_unit = TimeUnit('1h')
        
        date_time1 = dt(2015,10,25,0,15,0, tzinfo='Europe/Rome')
        date_time2 = date_time1 + time_unit    # 2015-10-25 01:15:00+02:00    
        date_time3 = date_time2 + time_unit    # 2015-10-25 02:15:00+02:00
        date_time4 = date_time3 + time_unit    # 2015-10-25 02:15:00+01:00

        date_time1_rounded = dt(2015,10,25,0,0,0, tzinfo='Europe/Rome')
        date_time2_rounded = date_time1_rounded + time_unit        
        date_time3_rounded = date_time2_rounded + time_unit
        date_time4_rounded = date_time3_rounded + time_unit
        date_time5_rounded = date_time4_rounded + time_unit
               
        self.assertEqual(time_unit.floor_dt(date_time2), date_time2_rounded)
        self.assertEqual(time_unit.ceil_dt(date_time2), date_time3_rounded)
          
        self.assertEqual(time_unit.floor_dt(date_time3), date_time3_rounded)
        self.assertEqual(time_unit.ceil_dt(date_time3), date_time4_rounded)
        
        self.assertEqual(time_unit.floor_dt(date_time4), date_time4_rounded)
        self.assertEqual(time_unit.ceil_dt(date_time4), date_time5_rounded)

        # Test ceil/floor/round with a calendar timeunit and across a DST change
        
        # Day unit
        time_unit = TimeUnit('1D')
        
        date_time1 = dt(2015,10,25,4,15,34, tzinfo='Europe/Rome') # DST off (+01:00)
        date_time1_floor = dt(2015,10,25,0,0,0, tzinfo='Europe/Rome') # DST on (+02:00)
        date_time1_ceil = dt(2015,10,26,0,0,0, tzinfo='Europe/Rome') # DST off (+01:00)

        self.assertEqual(time_unit.floor_dt(date_time1), date_time1_floor)
        self.assertEqual(time_unit.ceil_dt(date_time1), date_time1_ceil)

        # Month unit
        time_unit = TimeUnit('1M')
        
        date_time1 = dt(2015,10,25,4,15,34, tzinfo='Europe/Rome') # DST off (+01:00)
        date_time1_floor = dt(2015,10,1,0,0,0, tzinfo='Europe/Rome') # DST on (+02:00)
        date_time1_ceil = dt(2015,11,1,0,0,0, tzinfo='Europe/Rome') # DST off (+01:00)

        self.assertEqual(time_unit.floor_dt(date_time1), date_time1_floor)
        self.assertEqual(time_unit.ceil_dt(date_time1), date_time1_ceil)

        # Week unit
        # TODO: not implemented...

        # Year unit
        time_unit = TimeUnit('1Y')
        
        date_time1 = dt(2015,10,25,4,15,34, tzinfo='Europe/Rome')
        date_time1_floor = dt(2015,1,1,0,0,0, tzinfo='Europe/Rome')
        date_time1_ceil = dt(2016,1,1,0,0,0, tzinfo='Europe/Rome')

        self.assertEqual(time_unit.floor_dt(date_time1), date_time1_floor)
        self.assertEqual(time_unit.ceil_dt(date_time1), date_time1_ceil)


    def tearDown(self):
        pass




# -*- coding: utf-8 -*-
"""Series transformations as slotting and resampling."""

from .time import dt_from_s, s_from_dt
from .datastructures import DataTimeSlot, DataTimeSlotSeries, TimePoint, DataTimePointSeries, DataTimePoint
from .utilities import compute_data_loss
from .operations import avg
from .units import TimeUnit

# Setup logging
import logging
logger = logging.getLogger(__name__)



#==========================
#  Base Transformation
#==========================

class Transformation(object):
    """Base transformation class."""
    
    @classmethod
    def __str__(cls):
        return '{} transformation'.format(cls.__name__.replace('Operator',''))

    def process(self, *args, **kwargs):
        raise NotImplementedError('This transformation is not implemented.')



#==========================
#  Resampler Transformation
#==========================

class Resampler(Transformation):
    """Resampler transformation."""

    def __init__(self, unit, interpolation_method='linear'):
        if isinstance(unit, TimeUnit):
            self.time_unit = unit
        else:
            self.time_unit = TimeUnit(unit)
        if self.time_unit.is_calendar():
            raise ValueError('Sorry, calendar time units are not supported by the Resampler (got "{}"). Use the Slotter instead.'.format(self.time_unit))
        self.interpolation_method=interpolation_method

    def _compute_resampled_point(self, data_time_point_series, unit, start_t, end_t, validity, timezone, fill_with, force_data_loss, fill_gaps, series_indexes, series_resolution, first_last):

        # Compute data_loss        
        point_data_loss = compute_data_loss(data_time_point_series,
                                            from_t   = start_t,
                                            to_t     = end_t,
                                            series_resolution = series_resolution,
                                            validity = validity,
                                            first_last = first_last)
        
        # TODO: unroll the following before the compute resampled point call
        interval_timeseries = DataTimePointSeries()
        prev_point = None
        next_point = None
        for data_time_point in data_time_point_series:
            # TODO: better check the math here. We use >= and <= because, since we are basically comparing intervals,
            #       the "right" excluded rule is excluded by comparing right with left and then left with right.
            if (data_time_point.t+(validity/2)) <= start_t:
                prev_point = data_time_point
                continue
            if (data_time_point.t-(validity/2)) >= end_t:
                next_point = data_time_point
                continue
            interval_timeseries.append(data_time_point)

        # If we have to fully reconstruct data
        if not interval_timeseries: #point_data_loss == 1: 

            # Reconstruct (fill_gaps)
            slot_data = {}
            for key in data_time_point_series[0].data.keys():
                
                # Special case with only one datapoint (i.e. beginning or end)
                if len(data_time_point_series) == 1:
                    slot_data[key] = data_time_point_series[0].data[key]
                
                else:   
                
                    prev_point = data_time_point_series[0]
                    next_point = data_time_point_series[1]
                
                    if self.interpolation_method == 'linear': 
                        diff = next_point.data[key] - prev_point.data[key]
                        delta_t = next_point.t - prev_point.t
                        ratio = diff / delta_t
                        point_t = start_t + (unit.duration_s() /2)
                        slot_data[key] = prev_point.data[key] + ((point_t - prev_point.t) * ratio)
    
                    elif self.interpolation_method == 'uniform':
                        slot_data[key] = (prev_point.data[key] + next_point.data[key]) /2
                   
                    else:
                        raise Exception('Unknown interpolation method "{}"'.format(self.interpolation_method))

        else:

            # Keys shortcut
            keys = data_time_point_series.data_keys()

            logger.debug('Slot timeseries: %s', interval_timeseries)

            # Compute sample averages data
            avgs = avg(interval_timeseries, prev_point=prev_point, next_point=next_point)
                            
            # Do we have a 100% and a fill_with?
            if fill_with is not None and point_data_loss == 1:
                slot_data = {key:fill_with for key in keys}                
            else:

                if isinstance(avgs, dict):
                    slot_data = {key:avgs[key] for key in keys}
                else:
                    slot_data = {keys[0]: avgs}
                    

        # Do we have a force data_loss? #TODO: do not compute data_loss if fill_with not present and force_data_loss 
        if force_data_loss is not None:
            point_data_loss = force_data_loss
        
#         # Create the DataTimePoint if we have data_loss
#         if not point_data_loss:
#             data_time_point = None
#         else:      
#             pass

        # Create the data time point
        data_time_point = DataTimePoint(t = (start_t+((end_t-start_t)/2)),
                                        tz = timezone,
                                        data  = slot_data,
                                        data_loss = point_data_loss)


        # Now handle indexes
        for index in series_indexes:
            if interval_timeseries:
                if index == 'data_loss':
                    continue
                index_sum = 0
                index_count = 0
                for item in interval_timeseries:
                    
                    # Get index value
                    try:
                        index_value = getattr(item, index)
                    except:
                        new_point_index_value = None
                    else:
                        if index_value is not None:
                            index_count += 1
                            index_sum += index_value
        
                # Compute the new index value (if there were indexes not None)
                if index_count > 0:
                    new_point_index_value = index_sum/index_count
                else:
                    new_point_index_value = None
    
            else:
                new_point_index_value = None
                
            # Set the index. Handle special case for data_reconstructed
            if index == 'data_reconstructed':
                setattr(data_time_point, '_data_reconstructed', new_point_index_value)
            else:
                setattr(data_time_point, index, new_point_index_value)                

        # Return
        return data_time_point


    def process(self, data_time_point_series, from_t=None, to_t=None, from_dt=None, to_dt=None,
                 validity=None, force_close_last=True, include_extremes=False, fill_with=None,
                 force_data_loss=None, fill_gaps=True, force=False):
        """Start the resampling process. If start and/or end are not set, they are set automatically
        based on first and last points of the series"""

        if not isinstance(data_time_point_series, DataTimePointSeries):
            raise TypeError('Can process only DataTimePointSeries, got "{}"'.format(data_time_point_series.__class__.__name__))

        if not data_time_point_series:
            raise ValueError('Cannot process empty data_time_point_series')

        if include_extremes:
            if from_t is not None or to_t is not None:
                raise ValueError('Setting "include_extremes" is not compatible with giving a from_t or a to_t')
            from_rounding_method = 'floor'
            to_rounding_method   = 'ceil'        
        else:
            from_rounding_method = 'ceil'
            to_rounding_method   = 'floor'

        # Move fromt_dt and to_dt to epoch to simplify the following
        if from_dt is not None:
            from_t = s_from_dt(from_dt)
        if to_dt is not None:
            to_t = s_from_dt(to_dt)
            # Also force close if we have explicitly set an end
            force_close_last = True

        # Set "from" if not set, otherwise check for consistency # TODO: move to steaming
        if from_t is None:
            from_t = data_time_point_series[0].t
            from_dt = dt_from_s(from_t, tz=data_time_point_series.tz)
            # Is the point already rounded to the time unit or do we have to round it ourselves?
            if not from_dt == self.time_unit.round_dt(from_dt):
                from_dt = self.time_unit.round_dt(from_dt, how=from_rounding_method)
                from_t  = s_from_dt(from_dt)
        else:
            from_dt = dt_from_s(from_t, tz=data_time_point_series.tz)
            if from_dt != self.time_unit.round_dt(from_dt):
                raise ValueError('Sorry, provided from_t is not consistent with the self.time_unit of "{}" (Got "{}")'.format(self.time_unit, from_t))
        
        # Set "to" if not set, otherwise check for consistency # TODO: move to streaming
        if to_t is None:
            to_t = data_time_point_series[-1].t
            to_dt = dt_from_s(to_t, tz=data_time_point_series.tz)
            # Is the point already rounded to the time unit or do we have to round it ourselves?
            if not to_dt == self.time_unit.round_dt(to_dt):
                to_dt = self.time_unit.round_dt(to_dt, how=to_rounding_method)
                to_t  = s_from_dt(to_dt)
        else:
            to_dt = dt_from_s(to_t, tz=data_time_point_series.tz)
            if to_dt != self.time_unit.round_dt(to_dt):
                raise ValueError('Sorry, provided to_t is not consistent with the self.time_unit of "{}" (Got "{}")'.format(self.time_unit, to_t))

        # Move the start back of half and the end forward of
        # half unit as well, as the point will be in the center
        from_t = from_t - (self.time_unit.duration_s() /2)
        from_dt = dt_from_s(from_t, data_time_point_series.tz)
        to_t = to_t + (self.time_unit.duration_s() /2)
        to_dt = dt_from_s(to_t, data_time_point_series.tz)

        # Automatically detect validity if not set
        if validity is None:
            validity = data_time_point_series.autodetected_sampling_interval
            logger.info('Using auto-detected sampling interval: %ss', validity)

        # Check if not upsamplimg (with some tolearance):
        if not force:
            if validity > (self.time_unit.duration_s() * 1.10):
                raise ValueError('Upsampling not supported yet (resampler unit: {}; detected time series sampling interval: {})'.format(self.time_unit, validity))

        logger.debug('Started slotter from "%s" (%s) to "%s" (%s)', from_dt, from_t, to_dt, to_t)

        if from_dt >= to_dt:
            raise ValueError('Sorry, from is >= to! (from_t={}, to_t={})'.format(from_t, to_t))

        # Set some support vars
        slot_start_t       = None
        slot_end_t         = None
        prev_data_time_point = None
        working_serie      = DataTimePointSeries()
        process_ended      = False
        resampled_data_time_point_series = DataTimePointSeries()

        # Set timezone
        timezone  = data_time_point_series.tz
        logger.debug('Using timezone "%s"', timezone)
       
        # Counters
        count = 0
        first = True

        # Indexes
        series_indexes = data_time_point_series.indexes
        series_resolution = data_time_point_series.resolution

        # Now go trough all the data in the time series        
        for data_time_point in data_time_point_series:

            logger.debug('Processing %s', data_time_point)

            # Increase counter
            count += 1
            
            # Set start_dt if not already done TODO: implement it correctly
            #if not from_dt:
            #    from_dt = self.time_unit.timeInterval.round_dt(data_time_point.dt) if rounded else data_time_point.dt
            
            # Pretend there was a slot before if we are at the beginning. TOOD: improve me.
            if slot_end_t is None:   
                slot_end_t = from_t

            # First, check if we have some points to discard at the beginning       
            if data_time_point.t < from_t:
                # If we are here it means we are going data belonging to a previous slot
                # (probably just spare data loaded to have access to the prev_datapoint)  
                prev_data_time_point = data_time_point
                continue

            # Similar concept for the end
            # TODO: what if we are in streaming mode? add if to_t is not None?
            if data_time_point.t >= to_t:
                if process_ended:
                    continue

            # The following procedure works in general for slots at the beginning and in the middle.
            # The approach is to detect if the current slot is "outdated" and spin a new one if so.

            if data_time_point.t > slot_end_t:
                # If the current slot is outdated:
                         
                # 1) Add this last point to the data_time_point_series:
                working_serie.append(data_time_point)
                 
                #2) keep spinning new slots until the current data point falls in one of them.
                
                # NOTE: Read the following "while" more as an "if" which can also lead to spin multiple
                # slot if there are empty slots between the one being closed and the data_time_point.dt.
                # TODO: leave or remove the above if for code readability?
                
                while slot_end_t < data_time_point.t:
                    logger.debug('Checking for end {} with point {}'.format(slot_end_t, data_time_point.t))
                    # If we are in the pre-first slot, just silently spin a new slot:
                    if slot_start_t is not None:
                        
                        # Append last point. Can be appended to multiple slots, this is normal since
                        # the empty slots in the middle will have only a far prev and a far next.
                        # can also be appended several times if working_serie is not reset (looping in the while)
                        if data_time_point not in working_serie:
                            working_serie.append(data_time_point)
  
                        logger.debug('This slot (start={}, end={}) is closed, now aggregating it..'.format(slot_start_t, slot_end_t))
                        
                        logger.debug('working_serie len: %s', len(working_serie))
                        logger.debug('working_serie first point dt: %s', working_serie[0].dt)
                        logger.debug('working_serie  last point dt: %s', working_serie[-1].dt)

                        
                        # Compute slot...
                        dataTimePoint = self._compute_resampled_point(working_serie,
                                                                      unit     = self.time_unit,
                                                                      start_t  = slot_start_t,
                                                                      end_t    = slot_end_t,
                                                                      validity = validity,
                                                                      timezone = timezone,
                                                                      fill_with = fill_with,
                                                                      force_data_loss = force_data_loss,
                                                                      fill_gaps = fill_gaps,
                                                                      series_indexes = series_indexes,
                                                                      series_resolution = series_resolution,
                                                                      first_last = first)
                        # Set first to false
                        if first:
                            first = False
                        
                        # .. and append results
                        if dataTimePoint:
                            logger.debug('Computed datapoint: %s',dataTimePoint )
                            resampled_data_time_point_series.append(dataTimePoint)


                    # Create a new slot. This is where all the "conventional" time logic kicks-in, and where the time zone is required.
                    slot_start_t = slot_end_t
                    slot_end_t   = s_from_dt(dt_from_s(slot_start_t, tz=timezone) + self.time_unit)

                    # Create a new working_serie as part of the "create a new slot" procedure
                    working_serie = DataTimePointSeries()
                    
                    # Append the previous prev_data_time_point to the new DataTimeSeries
                    if prev_data_time_point:
                        working_serie.append(prev_data_time_point)

                    logger.debug('Spinned a new slot (start={}, end={})'.format(slot_start_t, slot_end_t))
                    
                # If last slot mark process as completed (and aggregate last slot if necessary)
                if data_time_point.dt >= to_dt:

                    # Edge case where we would otherwise miss the last slot
                    if data_time_point.dt == to_dt:
                        
                        # Compute slot...
                        dataTimePoint = self._compute_resampled_point(working_serie,
                                                                      unit     = self.time_unit, 
                                                                      start_t  = slot_start_t,
                                                                      end_t    = slot_end_t,
                                                                      validity = validity,
                                                                      timezone = timezone,
                                                                      fill_with = fill_with,
                                                                      force_data_loss = force_data_loss,
                                                                      fill_gaps = fill_gaps,
                                                                      series_indexes = series_indexes,
                                                                      series_resolution = series_resolution,
                                                                      first_last = True)
                        
                        # .. and append results
                        if dataTimePoint:
                            resampled_data_time_point_series.append(dataTimePoint)

                    process_ended = True
                    

            # Append this point to the working serie
            working_serie.append(data_time_point)
            
            # ..and save as previous point
            prev_data_time_point =  data_time_point           


        # Last slots
        if force_close_last:

            # 1) Close the last slot and aggreagte it. You should never do it unless you knwo what you are doing
            if working_serie:
    
                logger.debug('This resampled point (start={}, end={}) is done, now computing it..'.format(slot_start_t, slot_end_t))
      
                # Compute slot...
                dataTimePoint = self._compute_resampled_point(working_serie,
                                                              unit     = self.time_unit, 
                                                              start_t  = slot_start_t,
                                                              end_t    = slot_end_t,
                                                              validity = validity,
                                                              timezone = timezone,
                                                              fill_with = fill_with,
                                                              force_data_loss = force_data_loss,
                                                              fill_gaps = fill_gaps,
                                                              series_indexes = series_indexes,
                                                              series_resolution = series_resolution,
                                                              first_last = True)
                
                # .. and append results
                if dataTimePoint:
                    resampled_data_time_point_series.append(dataTimePoint)

            # 2) Handle missing slots until the requested end (end_dt)
            # TODO: Implement it. Sure?

        logger.info('Resampled %s DataTimePoints in %s DataTimePoints', count, len(resampled_data_time_point_series))

        return resampled_data_time_point_series



#==========================
#  Slotter Transformation
#==========================

class Slotter(Transformation):
    """Slotter transformation."""

    def __init__(self, unit, default_operation=avg, extra_operations=None, interpolation_method='linear'):
        if isinstance(unit, TimeUnit):
            self.time_unit = unit
        else:
            self.time_unit = TimeUnit(unit)
        self.default_operation=default_operation
        self.extra_operations=extra_operations
        self.interpolation_method=interpolation_method


    def _compute_slot(self, data_time_point_series, unit, start_t, end_t, validity, timezone, fill_with, force_data_loss, fill_gaps, series_indexes, series_resolution, first_last):

        # Compute data_loss
        slot_data_loss = compute_data_loss(data_time_point_series,
                                           from_t   = start_t,
                                           to_t     = end_t,
                                           series_resolution = series_resolution,
                                           validity = validity,
                                           first_last = first_last)
        
        # Initialize slot data
        slot_data = {}

        # TODO: unroll the following before the compute slot call
        slot_timeseries = DataTimePointSeries()
        prev_point = None
        next_point = None
        for data_time_point in data_time_point_series:
            if (data_time_point.t+(validity/2)) < start_t:
                prev_point = data_time_point
                continue
            if (data_time_point.t-(validity/2)) >= end_t:
                next_point = data_time_point
                continue
            slot_timeseries.append(data_time_point)


        # If we have to fully reconstruct data
        # TODO: we have a huge conceputal problem here if checkong on the slot_data_loss:
        #       if data_loss is 1 but due to point data loses (maybe even reconstructed)
        #       and not entirely missing points, we should actually compute as if not a full data loss.
        
        if not slot_timeseries: # slot_data_loss == 1: 

            # Reconstruct (fill gaps)
            for key in data_time_point_series[0].data.keys():
                
                # Set data as None, will be interpolated afterwards
                slot_data['{}_{}'.format(key, self.default_operation.__name__)] = None
                
                # Handle also extra ops
                if self.extra_operations:
                    for extra_operation in self.extra_operations:
                        slot_data['{}_{}'.format(key, extra_operation.__name__)] = None

        else:
 
            # Keys shortcut
            keys = data_time_point_series.data_keys()

            # Compute the default operation (in some cases it might not be defined, hence the "if")
            if self.default_operation:
                default_operation_data = self.default_operation(slot_timeseries, prev_point=prev_point, next_point=next_point)
                                
                # Do we have a 100% and a fill_with?
                if fill_with is not None and slot_data_loss == 1:
                    for key in keys:
                        slot_data['{}_{}'.format(key, self.default_operation.__name__)] = fill_with
                   
                else:
    
                    #if isinstance(default_operation_data, dict):
                    #    slot_data = {key:default_operation_data[key] for key in keys}
                    #else:
                    #    slot_data = {keys[0]: default_operation_data}
                    
                    if isinstance(default_operation_data, dict):
                        for key in keys:
                            slot_data['{}_{}'.format(key, self.default_operation.__name__)] = default_operation_data[key]
                    else:
                        slot_data['{}_{}'.format(keys[0], self.default_operation.__name__)] = default_operation_data

            # Handle extra operations
            if self.extra_operations:
                for extra_operation in self.extra_operations:
                    extra_operation_data = extra_operation(slot_timeseries, prev_point=prev_point, next_point=next_point)
                    if isinstance(extra_operation_data, dict):
                        for result_key in extra_operation_data:
                            slot_data['{}_{}'.format(result_key, extra_operation.__name__)] = extra_operation_data[result_key]
                    else:
                        slot_data['{}_{}'.format(keys[0], extra_operation.__name__)] = extra_operation_data

        # Do we have a force data_loss? #TODO: do not compute data_loss if fill_with not present and force_data_loss 
        if force_data_loss is not None:
            slot_data_loss = force_data_loss

        # Create the DataTimeSlot
        data_time_slot = DataTimeSlot(start = TimePoint(t=start_t, tz=timezone),
                                      end   = TimePoint(t=end_t, tz=timezone),
                                      unit  = unit,
                                      data  = slot_data,
                                      data_loss = slot_data_loss)

        # Now handle indexes
        for index in series_indexes:
            if slot_timeseries:
                if index == 'data_loss':
                    continue
                index_sum = 0
                index_count = 0
                for item in slot_timeseries:
                    
                    # Get index value
                    try:
                        index_value = getattr(item, index)
                    except:
                        slotted_index_value = None
                    else:
                        if index_value is not None:
                            index_count += 1
                            index_sum += index_value
        
                # Compute the slotted index value (if there were indexes not None)
                if index_count > 0:
                    slotted_index_value = index_sum/index_count
                else:
                    slotted_index_value = None
    
            else:
                slotted_index_value = None

            # Set the index. Handle special case for data_reconstructed
            if index == 'data_reconstructed':
                setattr(data_time_slot, '_data_reconstructed', slotted_index_value)
            else:
                setattr(data_time_slot, index, slotted_index_value)                
        
                    
        # Return the slot
        return data_time_slot


    def process(self, data_time_point_series, from_t=None, from_dt=None, to_t=None, to_dt=None, validity=None, force_close_last=False,
                 include_extremes=False, fill_with=None, force_data_loss=None, fill_gaps=True, force=False):
        """Start the slotting process. If start and/or end are not set, they are set automatically based on first and last points of the series."""

        if not isinstance(data_time_point_series, DataTimePointSeries):
            raise TypeError('Can process only DataTimePointSeries, got "{}"'.format(data_time_point_series.__class__.__name__))

        if not data_time_point_series:
            raise ValueError('Cannot process empty data_time_point_series')

        if include_extremes:
            if from_t is not None or to_t is not None:
                raise ValueError('Setting "include_extremes" is not compatible with giving a from_t or a to_t')
            from_rounding_method = 'floor'
            to_rounding_method   = 'ceil' 
            force_close_last = True       
        else:
            from_rounding_method = 'ceil'
            to_rounding_method   = 'floor'
            
        # Move fromt_dt and to_dt to epoch to simplify the following
        if from_dt is not None:
            from_t = s_from_dt(from_dt)
        if to_dt is not None:
            to_t = s_from_dt(to_dt)
            # Also force close if we have explicitly set an end
            force_close_last = True

        # Set "from" if not set, otherwise check for consistency # TODO: move to steaming
        if from_t is None:
            from_t = data_time_point_series[0].t
            from_dt = dt_from_s(from_t, data_time_point_series.tz)
            # Is the point already rounded to the time unit or do we have to round it ourselves?
            if not from_dt == self.time_unit.round_dt(from_dt):
                from_dt = self.time_unit.round_dt(from_dt, how=from_rounding_method)
                from_t  = s_from_dt(from_dt)
        else:
            from_dt = dt_from_s(from_t, data_time_point_series.tz)
            if from_dt != self.time_unit.round_dt(from_dt):
                raise ValueError('Sorry, provided from_t is not consistent with the self.time_unit of "{}" (Got "{}")'.format(self.time_unit, from_t))

        # Set "to" if not set, otherwise check for consistency # TODO: move to streaming
        if to_t is None:
            to_t = data_time_point_series[-1].t
            to_dt = dt_from_s(to_t, data_time_point_series.tz)
            # Is the point already rounded to the time unit or do we have to round it ourselves?
            if not to_dt == self.time_unit.round_dt(to_dt):
                to_dt = self.time_unit.round_dt(to_dt, how=to_rounding_method)
                to_t  = s_from_dt(to_dt)
        else:
            to_dt = dt_from_s(to_t, data_time_point_series.tz)
            if to_dt != self.time_unit.round_dt(to_dt):
                raise ValueError('Sorry, provided to_t is not consistent with the self.time_unit of "{}" (Got "{}")'.format(self.time_unit, to_t))
            # Also force close if we have explicitly set an end
            force_close_last = True
            
        # Automatically detect validity if not set
        if validity is None:
            validity = data_time_point_series.autodetected_sampling_interval
            logger.info('Using auto-detected sampling interval: %ss', validity)

        # Check if not upslotting (with some tolerance)
        if not force:
            # TODO: this check is super-weak. Will fail in loads of edge cases, i.e. months slotted in 30 days.
            unit_duration_s = self.time_unit.duration_s(data_time_point_series[0].dt)
            if validity > (unit_duration_s * 1.1):
                raise ValueError('Upslotting not supported yet (slotter unit: {}; detected time series sampling interval: {})'.format(unit_duration_s, validity))
            
        # Log
        logger.debug('Started slotter from "%s" (%s) to "%s" (%s)', from_dt, from_t, to_dt, to_t)

        if from_dt >= to_dt:
            raise ValueError('Sorry, from is >= to! (from_t={}, to_t={})'.format(from_t, to_t))

        # Set some support vars
        slot_start_t       = None
        slot_end_t         = None
        prev_data_time_point = None
        working_serie      = DataTimePointSeries()
        process_ended      = False
        data_time_slot_series  = DataTimeSlotSeries()
        slots_to_be_interpolated = []
        last_no_full_data_loss_slot = None

        # Set timezone
        timezone  = data_time_point_series.tz
        logger.debug('Using timezone "%s"', timezone)
       
        # Counters
        count = 0
        first = True

        # Indexes
        series_indexes = data_time_point_series.indexes
        series_resolution = data_time_point_series.resolution

        # Now go trough all the data in the time series        
        for data_time_point in data_time_point_series:

            logger.debug('Processing %s', data_time_point)

            # Increase counter
            count += 1
            
            # Set start_dt if not already done TODO: implement it correctly
            #if not from_dt:
            #    from_dt = self.time_unit.timeInterval.round_dt(data_time_point.dt) if rounded else data_time_point.dt
            
            # Pretend there was a slot before if we are at the beginning. TOOD: improve me.
            if slot_end_t is None:   
                slot_end_t = from_t

            # First, check if we have some points to discard at the beginning       
            if data_time_point.t < from_t:
                # If we are here it means we are going data belonging to a previous slot
                # (probably just spare data loaded to have access to the prev_datapoint)  
                prev_data_time_point = data_time_point
                continue

            # Similar concept for the end
            # TODO: what if we are in streaming mode? add if to_t is not None?
            if data_time_point.t >= to_t:
                if process_ended:
                    break

            # The following procedure works in general for slots at the beginning and in the middle.
            # The approach is to detect if the current slot is "outdated" and spin a new one if so.

            if data_time_point.t > slot_end_t:
                # If the current slot is outdated:
                         
                # 1) Add this last point to the data_time_point_series:
                working_serie.append(data_time_point)
                 
                #2) keep spinning new slots until the current data point falls in one of them.
                
                # NOTE: Read the following "while" more as an "if" which can also lead to spin multiple
                # slot if there are empty slots between the one being closed and the data_time_point.dt.
                # TODO: leave or remove the above if for code readability?
                
                while slot_end_t < data_time_point.t:
                    logger.debug('Checking for end {} with point {}'.format(slot_end_t, data_time_point.t))
                    # If we are in the pre-first slot, just silently spin a new slot:
                    if slot_start_t is not None:
                        
                        # Append last point. Can be appended to multiple slots, this is normal since
                        # the empty slots in the middle will have only a far prev and a far next.
                        # can also be appended several times if working_serie is not reset (looping in the while)
                        if data_time_point not in working_serie:
                            working_serie.append(data_time_point)
  
                        logger.debug('This slot (start={}, end={}) is closed, now aggregating it..'.format(slot_start_t, slot_end_t))
                        
                        logger.debug('working_serie first point dt: %s', working_serie[0].dt)
                        logger.debug('working_serie last point dt: %s', working_serie[-1].dt)

                        # Compute slot...
                        data_time_slot = self._compute_slot(working_serie,
                                                            unit     = self.time_unit,
                                                            start_t  = slot_start_t,
                                                            end_t    = slot_end_t,
                                                            validity = validity,
                                                            timezone = timezone,
                                                            fill_with = fill_with,
                                                            force_data_loss = force_data_loss,
                                                            fill_gaps = fill_gaps,
                                                            series_indexes = series_indexes,
                                                            series_resolution = series_resolution,
                                                            first_last = first)

                        # Set first to false
                        if first:
                            first = False
                        
                        # .. and append results (unless we are before the first timeseries start point)
                        if slot_end_t > data_time_point_series[0].t:
                            if data_time_slot.data_loss == 1.0:
                                # if data loss is full, append to slot to the slots to be interpolated
                                slots_to_be_interpolated.append(data_time_slot)
                            else:
                                # If we have slots to be intepolated
                                if slots_to_be_interpolated:
                                    for i, slot_to_be_interpolated in enumerate(slots_to_be_interpolated):

                                        # Prepare for interpolated data
                                        interpolated_data = {}

                                        # Computed interpolated data
                                        if self.interpolation_method == 'linear': 
                                            for data_key in data_time_slot_series.data_keys():                                                
                                                interpolated_data[data_key] = ((((data_time_slot.data[data_key] - last_no_full_data_loss_slot.data[data_key]) /
                                                                                 (len(slots_to_be_interpolated) + 1) ) * (i+1))  + last_no_full_data_loss_slot.data[data_key])
    
                                        elif self.interpolation_method == 'uniform':
                                            for data_key in data_time_slot_series.data_keys():                                                
                                                interpolated_data[data_key] = (((data_time_slot.data[data_key] - last_no_full_data_loss_slot.data[data_key]) / 2) 
                                                                                + last_no_full_data_loss_slot.data[data_key])
                                           
                                        else:
                                            raise Exception('Unknown interpolation method "{}"'.format(self.interpolation_method))

                                        # Add interpolated data
                                        slot_to_be_interpolated._data = interpolated_data
                                        data_time_slot_series.append(slot_to_be_interpolated)
                                        
                                    # Reset the "buffer"
                                    slots_to_be_interpolated = []

                                # Append this slot to the time series
                                data_time_slot_series.append(data_time_slot)
                                
                                # ... and set this slot as the last with no full data loss
                                last_no_full_data_loss_slot = data_time_slot


                    # Create a new slot. This is where all the "calendar" time unit logic kicks-in, and where the time zone is required.
                    slot_start_t = slot_end_t
                    slot_start_dt = dt_from_s(slot_start_t, tz=timezone)
                    slot_end_t   = s_from_dt(dt_from_s(slot_start_t, tz=timezone) + self.time_unit)
                    slot_end_dt = dt_from_s(slot_end_t, tz=timezone)

                    # Create a new working_serie as part of the "create a new slot" procedure
                    working_serie = DataTimePointSeries()
                    
                    # Append the previous prev_data_time_point to the new DataTimeSeries
                    if prev_data_time_point:
                        working_serie.append(prev_data_time_point)

                    logger.debug('Spinned a new slot (start={}, end={})'.format(slot_start_dt, slot_end_dt))
                    
                # If last slot mark process as completed (and aggregate last slot if necessary)
                if data_time_point.dt >= to_dt:

                    # Edge case where we would otherwise miss the last slot
                    if data_time_point.dt == to_dt:
                        
                        # Compute slot...
                        data_time_slot = self._compute_slot(working_serie,
                                                            unit     = self.time_unit, 
                                                            start_t  = slot_start_t,
                                                            end_t    = slot_end_t,
                                                            validity = validity,
                                                            timezone = timezone,
                                                            fill_with = fill_with,
                                                            force_data_loss = force_data_loss,
                                                            fill_gaps = fill_gaps,
                                                            series_indexes = series_indexes,
                                                            series_resolution = series_resolution,
                                                            first_last = True)
                        
                        # .. and append results 
                        data_time_slot_series.append(data_time_slot)

                    process_ended = True
                    

            # Append this point to the working serie
            working_serie.append(data_time_point)
            
            # ..and save as previous point
            prev_data_time_point =  data_time_point           

        # Last slots
        if force_close_last:

            # 1) Close the last slot and aggreagte it. You should never do it unless you knwo what you are doing
            if working_serie:
    
                logger.debug('This slot (start={}, end={}) is closed, now aggregating it..'.format(slot_start_t, slot_end_t))
      
                # Compute slot...
                data_time_slot = self._compute_slot(working_serie,
                                                    unit     = self.time_unit, 
                                                    start_t  = slot_start_t,
                                                    end_t    = slot_end_t,
                                                    validity = validity,
                                                    timezone = timezone,
                                                    fill_with = fill_with,
                                                    force_data_loss = force_data_loss,
                                                    fill_gaps = fill_gaps,
                                                    series_indexes = series_indexes,
                                                    series_resolution = series_resolution,
                                                    first_last = True)
                
                # .. and append results 
                data_time_slot_series.append(data_time_slot)

            # 2) Handle missing slots until the requested end (end_dt)
            # TODO: Implement it. Sure? Clashes with the idea of reconstructors..

        logger.info('Slotted %s DataTimePoints in %s DataTimeSlots', count, len(data_time_slot_series))

        return data_time_slot_series





# -*- coding: utf-8 -*-
"""Provides base model classes."""

import os
import json
import uuid
import statistics
from ..exceptions import NotFittedError
from ..utilities import check_timeseries, check_resolution, check_data_keys, item_is_in_range
from ..time import now_s, dt_from_s
from ..units import TimeUnit
from pandas import DataFrame
import shutil

# Keras
from keras.models import load_model as load_keras_model

# Setup logging
import logging
logger = logging.getLogger(__name__)

# Suppress TensorFlow warnings as default behavior
import tensorflow as tf
tf.compat.v1.logging.set_verbosity(tf.compat.v1.logging.ERROR)


#======================
#  Base Model
#======================

class Model(object):
    '''A stateless model, or a white-box model. Exposes only predict(), apply() and evaluate() methods,
    since it is assumed that all the information is coded and nothing is learnt from the data.'''
    
    def __init__(self):
        pass

    
    def predict(self, data, *args, **kwargs):
        """Call te model predict logic"""
        try:
            self._predict
        except AttributeError:
            raise NotImplementedError('Predicting from this model is not implemented')
        
        return self._predict(data, *args, **kwargs)


    def apply(self, data, *args, **kwargs):
        try:
            self._apply
        except AttributeError:
            raise NotImplementedError('Applying this model is not implemented')
        
        return self._apply(data, *args, **kwargs)


    def evaluate(self, data, *args, **kwargs):
        try:
            self._evaluate
        except AttributeError:
            raise NotImplementedError('Evaluating this model is not implemented')
        
        return self._evaluate(data, *args, **kwargs)



#=========================
#  Base parametric models
#=========================

class ParametricModel(Model):
    """A stateful model with parameters, or a (partly) black-box model. Parameters can be set manually
    or learnt (fitted) from data. On top of the predict(), apply() and evaluate() methods it provides
    a save() method to store the parameters of the model, and optionally a fit() method if the parameters
    are to be learnt form data."""
    
    def __init__(self, path=None, id=None):
        
        if path:
            with open(path+'/data.json', 'r') as f:
                self.data = json.loads(f.read())         
            self.fitted = True
            
        else:
            if not id:
                id = str(uuid.uuid4())
            self.fitted = False
            self.data = {'id': id}

        super(ParametricModel, self).__init__()


    def save(self, path):

        if not self.fitted:
            raise NotFittedError()

        # Prepare model dir and dump data as json
        model_dir = '{}/{}'.format(path, self.data['id'])
        os.makedirs(model_dir)
        model_data_file = '{}/data.json'.format(model_dir)
        with open(model_data_file, 'w') as f:
            f.write(json.dumps(self.data))
        logger.info('Saved model with id "%s" in "%s"', self.data['id'], model_dir)
        return model_dir

    
    def predict(self, data, *args, **kwargs):

        try:
            self._predict
        except AttributeError:
            raise NotImplementedError('Predicting from this model is not implemented')

        if not self.fitted:
            raise NotFittedError()

        return self._predict(data, *args, **kwargs)


    def apply(self, data, *args, **kwargs):

        try:
            self._apply
        except AttributeError:
            raise NotImplementedError('Applying this model is not implemented')

        if not self.fitted:
            raise NotFittedError()

        return self._apply(data, *args, **kwargs)


    def evaluate(self, data, *args, **kwargs):

        try:
            self._evaluate
        except AttributeError:
            raise NotImplementedError('Evaluating this model is not implemented')

        if not self.fitted:
            raise NotFittedError()
                
        return self._evaluate(data, *args, **kwargs)


    def fit(self, *args, **kwargs):

        try:
            self._fit
        except AttributeError:
            raise NotImplementedError('Fitting this model is not implemented')
        
        fit_output = self._fit(*args, **kwargs)

        self.data['fitted_at'] = now_s()
        self.fitted = True

        return fit_output


    def cross_validate(self, *args, **kwargs):

        try:
            self._cross_validate
        except AttributeError:
            raise NotImplementedError('Cross-validating this model is not implemented')

        return self._cross_validate(*args, **kwargs)


    @property
    def id(self):
        return self.data['id']



class TimeSeriesParametricModel(ParametricModel):
    """A parametric model specifically designed to work with time series data. In particular,
    it ensures resolution and data keys consistency between methods and save/load operations.""" 
    
    def __init__(self, path=None, id=None):

        # Call parent init
        super(TimeSeriesParametricModel, self).__init__(path, id)

        # If the model has been loaded, convert resolution as TimeUnit
        if self.fitted:
            self.data['resolution'] = TimeUnit(self.data['resolution'])

    def save(self, *args, **kwargs):

        # If fitted, handle resolution. If not fitted, the parent save will raise.
        if self.fitted:
        
            # Save original data resolution (Unit or TimeUnit Object)
            or_resolution = self.data['resolution']
    
            # Temporarily change the model resolution unit as string representation
            self.data['resolution'] = str(self.data['resolution'])
            
        # Call parent save
        save_output  = super(TimeSeriesParametricModel, self).save(*args, **kwargs)
        
        # Set back original data resolution
        self.data['resolution'] = or_resolution
        
        # Return output
        return save_output


    def fit(self, timeseries, *args, **kwargs):
        
        # Check timeseries
        check_timeseries(timeseries)
        
        # Call parent fit
        fit_output = super(TimeSeriesParametricModel, self).fit(timeseries, *args, **kwargs)
        
        # Set timeseries resolution
        self.data['resolution'] = timeseries.resolution
        
        # Set timeseries data keys
        self.data['data_keys'] = timeseries.data_keys()

        # Return output
        return fit_output


    def predict(self, timeseries, *args, **kwargs):
        
        # Check timeseries and its resolution
        check_timeseries(timeseries)
        
        # If fitted, check resolution and keys. If not fitted, the parent init will raise.
        if self.fitted:
            check_resolution(timeseries, self.data['resolution'])
            check_data_keys(timeseries, self.data['data_keys'])
                
        # Call parent predict and return output
        return super(TimeSeriesParametricModel, self).predict(timeseries, *args, **kwargs)


    def evaluate(self, timeseries, *args, **kwargs):
        
        # Check timeseries and its resolution
        check_timeseries(timeseries)
        
        # If fitted, check resolution. If not fitted, the parent init will raise.
        if self.fitted:
            check_resolution(timeseries, self.data['resolution'])
                
        # Call parent evaluate and return output
        return super(TimeSeriesParametricModel, self).evaluate(timeseries, *args, **kwargs)


    def apply(self, timeseries, *args, **kwargs):
        
        # Check timeseries
        check_timeseries(timeseries)
        
        # If fitted, check resolution. If not fitted, the parent init will raise.
        if self.fitted:
            check_resolution(timeseries, self.data['resolution'])
        
        # Call parent apply and return output
        return super(TimeSeriesParametricModel, self).apply(timeseries, *args, **kwargs)


    def _cross_validate(self, data, *args, **kwargs):

        try:
            self._evaluate
        except AttributeError:
            raise NotImplementedError('Evaluating this model is not implemented, cannot cross-validate')

        # Check timeseries
        check_timeseries(data)
        
        # If fitted, check resolution. If not fitted, the parent init will raise.
        if self.fitted:
            check_resolution(data, self.data['resolution'])
        
        # Decouple fit from validate args
        fit_kwargs = {}
        evaluate_kwargs = {}
        consumed_kwargs = []
        for kwarg in kwargs:
            if kwarg.startswith('fit_'):
                consumed_kwargs.append(kwarg)
                fit_kwargs[kwarg.replace('fit_', '')] = kwargs[kwarg]
            if kwarg.startswith('evaluate_'):
                consumed_kwargs.append(kwarg)
                evaluate_kwargs[kwarg.replace('evaluate_', '')] = kwargs[kwarg]
        for consumed_kwarg in consumed_kwargs:
            kwargs.pop(consumed_kwarg)

        # For readability
        timeseries = data
        
        # How many rounds
        rounds = kwargs.pop('rounds', 10)

        # Do we still have some kwargs?
        if kwargs:
            raise Exception('Got some unknown args: {}'.format(kwargs))
            
        # How many items per round?
        round_items = int(len(timeseries) / rounds)
        logger.debug('Items per round: {}'.format(round_items))
        
        # Start the fit / evaluate loop
        evaluations = []        
        for i in range(rounds):
            from_t = timeseries[(round_items*i)].t
            try:
                to_t = timeseries[(round_items*i) + round_items].t
            except IndexError:
                to_t = timeseries[(round_items*i) + round_items - 1].t
            from_dt = dt_from_s(from_t)
            to_dt   = dt_from_s(to_t)
            logger.info('Cross validation round #{} of {}: validate from {} ({}) to {} ({}), fit on the rest.'.format(i+1, rounds, from_t, from_dt, to_t, to_dt))
            
            # Fit
            if i == 0:            
                logger.debug('Fitting from {} ({})'.format(to_t, to_dt))
                self.fit(data, from_t=to_t, **fit_kwargs)
            else:
                logger.debug('Fitting until {} ({}) and then from {} ({}).'.format(to_t, to_dt, from_t, from_dt))
                self.fit(data, from_t=to_t, to_t=from_t, **fit_kwargs)                
            
            # Evaluate & append
            evaluations.append(self.evaluate(data, from_t=from_t, to_t=to_t, **evaluate_kwargs))
        
        # Regroup evaluations
        evaluation_metrics = list(evaluations[0].keys())
        scores_by_evaluation_metric = {}
        for evaluation in evaluations:
            for evaluation_metric in evaluation_metrics:
                try:
                    scores_by_evaluation_metric[evaluation_metric].append(evaluation[evaluation_metric])
                except KeyError:
                    try:
                        scores_by_evaluation_metric[evaluation_metric] = [evaluation[evaluation_metric]]
                    except KeyError:
                        raise Exception('Error, the model generated different evaluation metrics over the rounds, cannot compute cross validation.') from None
        
        # Prepare and return results
        results = {}
        for evaluation_metric in scores_by_evaluation_metric:
            results[evaluation_metric+'_avg'] = statistics.mean(scores_by_evaluation_metric[evaluation_metric])
            results[evaluation_metric+'_stdev'] = statistics.stdev(scores_by_evaluation_metric[evaluation_metric])         
        return results



#=========================
#  Base Prophet model
#=========================

class ProphetModel(TimeSeriesParametricModel):
    '''Class to wrap some common logic for Prophet-based models.'''


    @classmethod
    def remove_timezone(cls, dt):
        return dt.replace(tzinfo=None)

    @classmethod
    def from_timeseria_to_prophet(cls, timeseries, from_t=None, to_t=None):

        # Create Python lists with data
        try:
            timeseries[0].data[0]
            data_keys_are_indexes = True
        except KeyError:
            timeseries[0].data.keys()
            data_keys_are_indexes = False
        
        data_as_list=[]
        for item in timeseries:
            
            # Skip if needed
            try:
                if not item_is_in_range(item, from_t, to_t):
                    continue
            except StopIteration:
                break                

            if data_keys_are_indexes:     
                data_as_list.append([cls.remove_timezone(item.dt), item.data[0]])
            else:
                data_as_list.append([cls.remove_timezone(item.dt), item.data[list(item.data.keys())[0]]])

        # Create the pandas DataFrames
        data = DataFrame(data_as_list, columns = ['ds', 'y'])

        return data



#=========================
#  Base ARIMA model
#=========================

class ARIMAModel():
    '''Class to wrap some common logic for ARIMA-based models.'''
    
    def get_start_end_indexes(self, timeseries, n):

        # Do the math to get the right indexes for the prediction, both out-of-sample and in-sample
        # Not used at the moment as there seems to be bugs in the statsmodel package.
        # See https://www.statsmodels.org/devel/generated/statsmodels.tsa.arima_model.ARIMAResults.predict.html
        
        # The default start_index for an arima prediction id the inxed after the fit timeseries
        # I.e. if the fit timeseries had 101 lements, the start index is the 101
        # So we need to compute the start and end indexes according to the fit timeseries,
        # we will use the "resolution" for this.
        
        # Save the requested prediction "from"
        requested_predicting_from_dt = timeseries[-1].dt + timeseries.resolution

        # Search for the index number TODO: try first with "pure" math which would work for UTC
        slider_dt = self.fit_timeseries[0].dt
        count = 0
        while True:
            if slider_dt  == requested_predicting_from_dt:
                break
             
            # To next item index
            if requested_predicting_from_dt < self.fit_timeseries[0].dt:
                slider_dt = slider_dt - self.fit_timeseries.resolution
                if slider_dt < requested_predicting_from_dt:
                    raise Exception('Miss!')
 
            else:
                slider_dt = slider_dt + self.fit_timeseries.resolution
                if slider_dt > requested_predicting_from_dt:
                    raise Exception('Miss!')
            
            count += 1
        
        # TODO: better understand the indexes in statsmodels (and potentially their bugs), as sometimes
        # setting the same start and end gave two predictions which is just nonsense in any scenario.
        start_index = count  
        end_index = count + n

        return (start_index, end_index)



#=========================
#  Base Keras model
#=========================

class KerasModel(ParametricModel):
    '''Class to wrap some common logic for Keras-based models.'''


    def __init__(self, path=None, id=None):
        
        super(KerasModel, self).__init__(path=path, id=id)

        # If loaded, load the Keras model as well
        if path:
            self.keras_model = load_keras_model('{}/keras_model'.format(path))


    def save(self, path):

        # Save the parameters (the "data") property
        model_dir = super(KerasModel, self).save(path)

        # Now save the Keras model itself
        try:
            self.keras_model.save('{}/keras_model'.format(model_dir))
        except Exception as e:
            shutil.rmtree(model_dir)
            raise e
        return model_dir


    @staticmethod
    def to_windows(timeseries):
        '''Compute window data values from a time series.'''
        key = timeseries.data_keys()[0]
        window_data_values = []
        for item in timeseries:
            window_data_values.append(item.data[key])
        return window_data_values

        
    @staticmethod
    def to_window_datapoints_matrix(timeseries, window, forecast_n, encoder=None):
        '''Compute window datapoints matrix from a time series.'''
    
        window_datapoints = []
        for i, _ in enumerate(timeseries):
            if i <  window:
                continue
            if i == len(timeseries)-forecast_n:
                break
                    
            # Add window values
            row = []
            for j in range(window):
                row.append(timeseries[i-window+j])
            window_datapoints.append(row)
                
        return window_datapoints

    
    @staticmethod
    def to_target_values_vector(timeseries, window, forecast_n):
        '''Compute target values vector from a time series.'''
    
        data_keys = timeseries.data_keys()
    
        targets = []
        for i, _ in enumerate(timeseries):
            if i <  window:
                continue
            if i == len(timeseries)-forecast_n:
                break
            
            # Add forecast target value(s)
            row = []
            for j in range(forecast_n):
                for data_key in data_keys:
                    row.append(timeseries[i+j].data[data_key])
            targets.append(row)

        return targets


    @staticmethod
    def compute_window_features(window_datapoints, data_keys, features):
        """Compute features from a list of window data points (or slots).
        
        Args:
            window_datapoints (list): The list with the data points (or slots)
            data_keys(dict): the keys of the point (or slot) data.
            features(list): the list of the features to compute.
                Supported values are:
                ``values`` (use the data values), 
                ``diffs``  (use the diffs between the values), and 
                ``hours``  (use the hours of the timestamp).
        """

        available_features = ['values', 'diffs', 'hours']
        for feature in features:
            if feature not in available_features:
                raise ValueError('Unknown feature "{}"'.format(feature))
        
        window_features=[]

        for i in range(len(window_datapoints)):
            
            datapoint_features = []
            
            # 1) datapoint values (for all keys)
            if 'values' in features:
                for data_key in data_keys:
                    datapoint_features.append(window_datapoints[i].data[data_key])
                            
            # 2) Compute diffs on normalized datapoints
            if 'diffs' in features:
                for data_key in data_keys:
                    if i == 0:
                        diff = window_datapoints[1].data[data_key] - window_datapoints[0].data[data_key]
                    elif i == len(window_datapoints)-1:
                        diff = window_datapoints[-1].data[data_key] - window_datapoints[-2].data[data_key]
                    else:
                        diff = (window_datapoints[i+1].data[data_key] - window_datapoints[i-1].data[data_key]) /2
                    if diff == 0:
                        diff = 1
                    datapoint_features.append(diff)

            # 3) Hour (normlized)
            if 'hours' in features:
                datapoint_features.append(window_datapoints[i].dt.hour/24)
                
            # Now append to the window features
            window_features.append(datapoint_features)

        return window_features


# ModelBasedModel
# WindowBasedModel
# ModelBasedReconstructor
# WindowBasedReconstrutor








# -*- coding: utf-8 -*-
"""Provides base data structures: Points, Slots and Series with all their specializations."""

from .time import s_from_dt , dt_from_s, UTC, timezonize
from .units import Unit, TimeUnit
from .utilities import is_close, to_time_unit_string
from copy import deepcopy
from pandas import DataFrame
from datetime import datetime
from .exceptions import ConsistencyException

# Setup logging
import logging
logger = logging.getLogger(__name__)


#======================
#  Generic Series
#======================

class Series(list):
    '''A list of items coming one after another where every item is guaranteed to be of the same type and in an order or succession.'''

    # 1,2,3 are in order. 1,3,8 are in order. 1,8,3 re not in order.
    # 5,6,7 are integer succession. 5.3, 5.4, 5.5 are too in a succesisons. 5,6,8 are not in a succession. 
    # a group or a number of related or similar things, events, etc., arranged or occurring in temporal, spatial, or other order or succession; sequence.

    # By default the type is not defined
    __TYPE__ = None
    
    def __init__(self, *args, **kwargs):

        #if 'accept_None' in kwargs and kwargs['accept_None']:
        #    self.accept_None = True
        #else:
        #    self.accept_None = False

        for arg in args:
            self.append(arg)
            
        self._title = None

    def append(self, item):
        """Append an item to the series"""
        # logger.debug('Checking %s', item)
        
        # Set type if not already done
        if not self.__TYPE__:
            self.__TYPE__ = type(item)

        # Check type
        if not isinstance(item, self.__TYPE__):
            raise TypeError('Got incompatible type "{}", can only accept "{}"'.format(item.__class__.__name__, self.__TYPE__.__name__))
        
        # Check order or succession
        try:
            try:
                if not item.__succedes__(self[-1]):
                    raise ValueError('Not in succession ("{}" does not succeeds "{}")'.format(item,self[-1])) from None                    
            except IndexError:
                raise
            except AttributeError:
                try:
                    if not item > self[-1]:
                        raise ValueError('Not in order ("{}" does not follow "{}")'.format(item,self[-1])) from None
                except TypeError:
                    raise TypeError('Object of class "{}" does not implement a "__gt__" or a "__succedes__" method, cannot append it to a Series (which is ordered)'.format(item.__class__.__name__)) from None
        except IndexError:
            pass
        
        # Append
        super(Series, self).append(item)
    
    def __sum__(self, other):
        raise NotImplementedError
    
    @property
    def title(self):
        if self._title:
            return self._title
        else:
            return None

    @title.setter
    def title(self, title):
        self._title=title

    def __repr__(self):
        return '{} of #{} elements'.format(self.__class__.__name__, len(self))
    
    def __str__(self):
        return self.__repr__()
        
    # Python slice (i.e [0:35])
    def __getitem__(self, key):
        if isinstance(key, slice):   
            # TODO: improve and implement shallow copies please.         
            indices = range(*key.indices(len(self)))
            series = self.__class__()
            for i in indices:
                series.append(super(Series, self).__getitem__(i))
            try:
                series.mark = self.mark
            except:
                pass
            return series
        elif isinstance(key, str):   
            # Try filtering on this data key only
            return self.filter(key)
        else:
            return super(Series, self).__getitem__(key)


    # Operations
    def duplicate(self):
        return deepcopy(self)

    def filter(self, *args, **kwargs):
        from .operations import filter as filter_operation
        return filter_operation(self, *args, **kwargs) 

    def merge(self, *args, **kwargs):
        from .operations import merge as merge_operation
        return merge_operation(self, *args, **kwargs)

    def select(self, *args, **kwargs):
        from .operations import select as select_operation
        return select_operation(self, *args, **kwargs)
    
    def min(self, *args, **kwargs):
        from .operations import min as min_operation
        return min_operation(self, *args, **kwargs)

    def max(self, *args, **kwargs):
        from .operations import max as max_operation
        return max_operation(self, *args, **kwargs)    

    def avg(self, *args, **kwargs):
        from .operations import avg as avg_operation
        return avg_operation(self, *args, **kwargs)   

    def sum(self, *args, **kwargs):
        from .operations import sum as sum_operation
        return sum_operation(self, *args, **kwargs)  

    def derivative(self, *args, **kwargs):
        from .operations import derivative as derivative_operation
        return derivative_operation(self, *args, **kwargs)   

    def integral(self, *args, **kwargs):
        from .operations import integral as integral_operation
        return integral_operation(self, *args, **kwargs)   

    def diff(self, *args, **kwargs):
        from .operations import diff as diff_operation
        return diff_operation(self, *args, **kwargs)   

    def csum(self, *args, **kwargs):
        from .operations import csum as csum_operation
        return csum_operation(self, *args, **kwargs)
     
    def mavg(self, *args, **kwargs):
        from .operations import mavg as mavg_operation
        return mavg_operation(self, *args, **kwargs)  

    # Transformation-based operations
    def slot(self, *args, **kwargs):
        from .transformations import Slotter
        slotter = Slotter(*args, **kwargs)
        return slotter.process(self)  

    # Transformation-based operations
    def resample(self, *args, **kwargs):
        from .transformations import Resampler
        resampler = Resampler(*args, **kwargs)
        return resampler.process(self)  

    # Indexes
    @property
    def indexes(self):
        # TODO: this approach does not allow for custom-defined indexes.
        #       Maybe move indexes from item attributes to a dictionary?
        if not self:
            return []
        else:
            _indexes = []

            # Do we have data reconstructed indexes?
            try:
                for item in self:
                    if item.data_reconstructed is not None:
                        _indexes.append('data_reconstructed')
                        break
            except AttributeError:
                pass

            # Do we have data loss indexes?
            try:
                for item in self:
                    if item.data_loss is not None:
                        _indexes.append('data_loss')
                        break
            except AttributeError:
                pass

            # Do we have an anomaly index?
            try:
                for item in self:
                    if item.anomaly is not None:
                        _indexes.append('anomaly')
                        break
            except AttributeError:
                pass

            # Do we have a forecast index?
            try:
                for item in self:
                    if item.forecast is not None:
                        _indexes.append('forecast')
                        break
            except AttributeError:
                pass
            
            return _indexes

    @property
    def mark(self):
        try:
            return self._mark
        except AttributeError:
            return None

    @mark.setter
    def mark(self, value):
        # Check valid mark
        if not isinstance(value, (list, tuple)):
            raise TypeError('Series mark must be a list or tuple')
        if not len(value) == 2:
            raise ValueError('Series mark must be a list or tuple of two elements')    
        self._mark = value


    # Inherited methods to be edited
    def pop(self, i=None):
        """
        Remove the item at the end of the series, and return it. The `i` index is kept for interface
        compatibility, but if specified must be the index corresponding to the last element.
        """  
        if i and i != len(self):
            raise NotImplementedError('Cannot pop an item in the middle')
        super(Series, self).pop(i)

    
    def clear(self):
        """Remove all items from the series. Equivalent to del a[:]."""
        super(Series, self).clear()

        
    def index(self, x, start=None, end=None):
        """
        Return zero-based index in the series of the item whose value is equal to x.
        Raises a ValueError if there is no such item.

        The optional arguments start and end are interpreted as in the slice notation and are
        used to limit the search to a particular subsequence of the list. The returned index is
        computed relative to the beginning of the full sequence rather than the start argument.
        """
        super(Series, self).index(x, start, end)


    def copy(self):
        """Return a shallow copy of the list."""
        

    # Inherited methods to be disabled

    def remove(self):
        """Disabled (removing is not compatible with a succession)."""

    def insert(self, i, x):
        """Disabled (inserting is not compatible with a succession)."""

    def extend(self):
        """Disabled, use the `merge` instead."""

    def count(self, x):
        """Disabled (there is only one item)."""

    def sort(self, key=None, reverse=False):
        """Disabled (sorting is already guaranteed)."""

    def reverse(self):
        """Disabled (reversing is not compatible with an ordering)."""


#======================
#  Points
#======================

class Point():
    """A generic point."""

    def __init__(self, *args):
        if not args:
            raise Exception('A Point requires at least one coordinate, got none.')

        # Validate
        for arg in args:
            try:
                float(arg)
            except:
                raise Exception('Got non-numerical argument: "{}"'.format(arg))

        # Store (as tuple)
        self.coordinates = tuple(args)

    @property
    def __coordinates_repr__(self):
        #return ','.join(str(coordinate) for coordinate in self.coordinates)
        if len(self.coordinates)==1:
            return str(self.coordinates[0])
        else:
            return str(self.coordinates)

    def __repr__(self):
        return '{} @ {}'.format(self.__class__.__name__, self.__coordinates_repr__)

    def __str__(self):
        return self.__repr__()

    def __eq__(self, other):
        if not self.coordinates == other.coordinates:
            return False
        for i in range(0, len(self.coordinates)):
            if self.coordinates[i] != other.coordinates[i]:
                return False
        return True
    
    def __getitem__(self,index):
        return self.coordinates[index]


class TimePoint(Point):
    """A point in the time dimension."""
    
    def __init__(self, *args, **kwargs):

        # Handle time zone if any (removing it from kwargs)
        tz = kwargs.pop('tz', None)
        if tz:
            self._tz = timezonize(tz)

        # Do we have to handle kwargs?
        if kwargs:
            if 't' in kwargs:
                
                # Ok, will create the point in the standard way, just used the "t" kwarg to set the time coordinate 
                t = kwargs['t']
            
            elif 'dt' in kwargs:
                
                # Ok, will convert the datetime to epoch and then create the point in the standard way
                t = s_from_dt(kwargs['dt'])
                
                # If we do not have a time zone, can we use the one from the dt used to initialize this TimePoint?
                try:
                    self._tz
                except AttributeError:
                    if kwargs['dt'].tzinfo:
                        
                        #Do not set it if it is UTC, it is the default
                        if kwargs['dt'].tzinfo == UTC:
                            pass
                        else:
                            self._tz = kwargs['dt'].tzinfo
                            #raise NotImplementedError('Not yet tz from dt ("{}")'.format(kwargs['dt']))
                
            #else:
            #    raise Exception('Don\'t know how to handle all kwargs (got "{}")'.format(kwargs))

        # Cast or create in the standard way
        elif args:
            if len(args) > 1:
                raise Exception('Don\'t know how to handle all args (got "{}")'.format(args))
            
            if isinstance(args[0], TimePoint):
                t   = args[0].t
                self._tz = args[0].tz
            else:
                t = args[0]
            
        # Call parent init
        super(TimePoint, self).__init__(t)

    @property
    def t(self):
        return self.coordinates[0]
    
    def __gt__(self, other):
        if self.t > other.t:
            return True
        else:
            return False

    @property
    def tz(self):
        try:
            return self._tz
        except AttributeError:
            return UTC
    
    def change_timezone(self, new_timezone):
        self._tz = timezonize(new_timezone)

    @property
    def dt(self):
        return dt_from_s(self.t, tz=self.tz)

    def __repr__(self):
        return '{} @ {} ({})'.format(self.__class__.__name__, self.t, self.dt)
        # return '{} @ t={} ({})'.format(self.__class__.__name__, self.t, self.dt)
    


class DataPoint(Point):
    """A point that carries some data."""
    
    def __init__(self, *args, **kwargs):
        try:
            self._data = kwargs.pop('data')
        except KeyError:
            raise Exception('A DataPoint requires a special "data" argument (got only "{}")'.format(kwargs))
        
        try:
            data_loss = kwargs.pop('data_loss')
        except KeyError:
            pass
        else:
            self._data_loss = data_loss
        
        super(DataPoint, self).__init__(*args, **kwargs)

    def __repr__(self):
        try:
            return '{} with data "{}" and data_loss="{}"'.format(super(DataPoint, self).__repr__(), self.data, self._data_loss)            
        except:
            return '{} with data "{}"'.format(super(DataPoint, self).__repr__(), self.data)
    
    def __eq__(self, other):
        if self._data != other._data:
            return False
        return super(DataPoint, self).__eq__(other)

    @property
    def data(self):
        return self._data

    @property
    def data_loss(self):
        try:
            return self._data_loss
        except AttributeError:
            return None

    @property
    def data_reconstructed(self):
        try:
            return self._data_reconstructed
        except AttributeError:
            return None

    @data_reconstructed.setter
    def data_reconstructed(self, value):
        self._data_reconstructed = value



class DataTimePoint(DataPoint, TimePoint):
    """A point that carries some data in the time dimension."""
    pass

    # NOTE: the __repr__ used is from the DataPoint above, which in turn uses the TimePoint one.


#======================
#  Point Series
#======================

class PointSeries(Series):
    """A series of generic points, where each item is guaranteed to be ordered."""
    __TYPE__ = Point


class TimePointSeries(PointSeries):
    """A series of points in time, where each item is guaranteed to be ordered."""

    __TYPE__ = TimePoint

    def __init__(self, *args, **kwargs):

        tz = kwargs.pop('tz', None)
        if tz:
            self._tz = timezonize(tz)
        
        self._resolution = None

        super(TimePointSeries, self).__init__(*args, **kwargs)
        

    # Check time ordering and resolution
    def append(self, item):
        try:
            
            # This is to support the deepcopy, otherwise the original prev_t will be used
            if len(self)>0:
                
                # logger.debug('Checking time ordering for t="%s" (prev_t="%s")', item.t, self.prev_t)
                if item.t < self.prev_t:
                    raise ValueError('Time t="{}" is out of order (prev t="{}")'.format(item.t, self.prev_t))
                
                if item.t == self.prev_t:
                    raise ValueError('Time t="{}" is a duplicate'.format(item.t))
                
                if self._resolution is None:
                    
                    self._resolution = TimeUnit(to_time_unit_string(item.t - self.prev_t, friendlier=True))
                    
                elif self._resolution != 'variable':
                    if self._resolution != item.t - self.prev_t:
                        self._resolution = 'variable'

            self.prev_t = item.t
                
        except AttributeError:
            self.prev_t = item.t
       
        super(TimePointSeries, self).append(item)

    @property
    def tz(self):
        # Note: we compute the tz on the fly beacuse for point time series we assume to use the tz
        # attribute way lass than the slot time series, where the tz is instead computed at append-time.
        try:
            return self._tz
        except AttributeError:
            # Detect time zone on the fly
            detected_tz = None
            for item in self:
                if not detected_tz:
                    detected_tz = item.tz
                else:
                    # Terrible but seems like no other way to compare pytz.tzfile.* classes
                    if str(item.tz) != str(detected_tz): 
                        return UTC
            return detected_tz
    
    @tz.setter
    def tz(self, value):
        self._tz = timezonize(value) 

    def change_timezone(self, new_timezone):
        for time_point in self:
            time_point.change_timezone(new_timezone)
        self.tz = time_point.tz

    @property
    def resolution(self):
        return self._resolution

    

class DataPointSeries(PointSeries):
    """A series of data points, where each item is guaranteed to be ordered and to carry the same data type."""

    __TYPE__ = DataPoint

    # Check data compatibility
    def append(self, item):
        try:
            # logger.debug('Checking data compatibility: %s ', item.data)
            #if item.data is None and self.accept_None:
            #    pass
            #else:
            if not type(self.item_data_reference) == type(item.data):
                raise TypeError('Got different data: {} vs {}'.format(self.item_data_reference.__class__.__name__, item.data.__class__.__name__))
            if isinstance(self.item_data_reference, list):
                if len(self.item_data_reference) != len(item.data):
                    raise ValueError('Got different data lengths: {} vs {}'.format(len(self.item_data_reference), len(item.data)))
            if isinstance(self.item_data_reference, dict):
                if set(self.item_data_reference.keys()) != set(item.data.keys()):
                    raise ValueError('Got different data keys: {} vs {}'.format(self.item_data_reference.keys(), item.data.keys()))
            
        except AttributeError:
            # logger.debug('Setting data reference: %s', item.data)
            self.item_data_reference = item.data
            
        super(DataPointSeries, self).append(item)

    def data_keys(self):
        if len(self) == 0:
            return None
        else:
            # TODO: can we optimize here? Computing them once and then serving them does not work if someone changes data keys...
            try:
                return sorted(list(self[0].data.keys()))
            except AttributeError:
                return list(range(len(self[0].data)))

    def remove_data_loss(self):
        new_series = self.duplicate()
        for item in new_series:
            item._data_loss = None
        return new_series



class DataTimePointSeries(DataPointSeries, TimePointSeries):
    """A series of data points in time, where each item is guaranteed to be ordered and to carry the same data type."""

    __TYPE__ = DataTimePoint

    def __init__(self, *args, **kwargs):

        # Handle df kwarg
        df = kwargs.pop('df', None)
        
        # Handle first argument as dataframe
        if args and (isinstance(args[0], DataFrame)):
            df = args[0]

        # Do we have to initialize the time series from a dataframe?        
        if df is not None:

            # Create data time points list
            data_time_points = []
            
            # Get data frame labels
            labels = list(df.columns) 
            
            for row in df.iterrows():
                
                # Set the timestamp
                if not isinstance(row[0], datetime):
                    raise TypeError('A DataFrame with a DateTime index column is required')
                dt = row[0]
                
                # Prepare data
                data = {}
                for i,label in enumerate(labels):
                    data[label] = row[1][i]
                
                data_time_points.append(DataTimePoint(dt=dt, data=data))
            
            # Set the list of data time points
            super(DataTimePointSeries, self).__init__(*data_time_points, **kwargs)
            return None
        
        # Original init
        super(DataTimePointSeries, self).__init__(*args, **kwargs)

    @property
    def df(self):
        """The time series as a Pandas DataFrame object"""

        data_keys = self.data_keys()
        
        if self[0].data_loss is not None:
            dump_data_loss = True
        else:
            dump_data_loss = False
        
        if dump_data_loss:
            columns = ['Timestamp'] + data_keys + ['data_loss']
        else:
            columns = ['Timestamp'] + data_keys
            
        df = DataFrame(columns=columns)
        for item in self:
            values = [item.data[key] for key in data_keys]
            if dump_data_loss:
                df = df.append(DataFrame([[item.dt]+values+[item.data_loss]], columns=columns))
            else:
                df = df.append(DataFrame([[item.dt]+values], columns=columns))                
        df = df.set_index('Timestamp')
        return df
        
    def plot(self, engine='dg', *args, **kwargs):
        if engine=='mp':
            from .plots import matplotlib_plot
            matplotlib_plot(self, *args, **kwargs)
        elif engine=='dg':
            from .plots import dygraphs_plot
            dygraphs_plot(self, *args, **kwargs)
        else:
            raise ValueError('Unknown plotting engine "{}'.format(engine))
        
    @property
    def autodetected_sampling_interval(self):
        try:
            return self._autodetected_sampling_interval
        except AttributeError:
            from .utilities import detect_sampling_interval
            self._autodetected_sampling_interval = detect_sampling_interval(self)
            return self._autodetected_sampling_interval
    
    @property
    def _resolution_string(self):
        if isinstance(self.resolution, Unit):
            # Includes TimeUnits
            _resolution_string = '{} resolution'.format(self.resolution)
        else:
            autodetected_sampling_interval_as_str = str(self.autodetected_sampling_interval)
            if autodetected_sampling_interval_as_str.endswith('.0'):
                # TODO: use a friendlier resolution here as well, as above?
                autodetected_sampling_interval_as_str = autodetected_sampling_interval_as_str[:-2] 
            _resolution_string = 'variable resolution (~{}s)'.format(autodetected_sampling_interval_as_str)
        return _resolution_string

    def __repr__(self):
        if len(self):
            return 'Time series of #{} points at {}, from point @ {} ({}) to point @ {} ({})'.format(len(self), self._resolution_string, self[0].t, self[0].dt, self[-1].t, self[-1].dt)
        else:
            return 'Time series of #0 points'


#======================
#  Slots
#======================

class Slot():
    """A generic slot."""

    __POINT_TYPE__ = Point
    
    def __init__(self, start, end=None, unit=None):
        
        if not isinstance(start, self.__POINT_TYPE__):
            raise TypeError('Slot start must be a Point object (got "{}")'.format(start.__class__.__name__))
        
        # Instantiate unit if not already done
        if unit and not isinstance(unit, Unit):
            unit = Unit(unit)
        
        if end is None and unit is not None:
            if len(start.coordinates)>1:
                raise Exception('Sorry, setting a start and a unit only works in unidimensional spaces')
            
            # Handle start + unit
            end = start + unit
             
        if not isinstance(end, self.__POINT_TYPE__):
            raise TypeError('Slot end must be a Point object (got "{}")'.format(end.__class__.__name__))
        
        # TODO: remove the following check, or make it optional (i.e. not used by TimeSlots)?
        if len(start.coordinates) != len(end.coordinates):
            raise ValueError('Slot start and end dimensions must be the same (got "{}" vs "{}")'.format(start.coordinates, end.coordinates))
        if start == end:
            raise ValueError('{} start and end must not be the same (got start="{}", end="{}")'.format(self.__class__.__name__, start,end))

        self.start = start
        self.end   = end
        
        if unit is not None:
            self._unit = unit

    def __repr__(self):
        return '{} @ [{},{}]'.format(self.__class__.__name__, self.start.__coordinates_repr__, self.end.__coordinates_repr__)
    
    def __str__(self):
        return self.__repr__()

    def __eq__(self, other):
        if not self.start == other.start:
            return False
        if not self.end == other.end:
            return False
        return True

    def __succedes__(self, other):
        if other.end != self.start:
            return False
        else:
            return True

    @classmethod
    def _compute_length(cls, start, end):
        values = []
        for i in range(len(start.coordinates)):
            values.append(end.coordinates[i] - start.coordinates[i])
        
        return sum(values)/len(values)
            
    @property
    def length(self):
        try:
            self._length
        except AttributeError:
            self._length = self._compute_length(self.start, self.end)
        return self._length

    @property
    def unit(self):
        try:
            return self._unit
        except AttributeError:
            if len(self.start.coordinates) == 1:
                return Unit(self.end.coordinates[0] - self.start.coordinates[0])
            else:
                values = []
                for i in range(len(self.start.coordinates)):
                    values.append(self.end.coordinates[i] - self.start.coordinates[i])
                
                return Unit(values)


class TimeSlot(Slot):
    """A slot in the time dimension."""

    __POINT_TYPE__ = TimePoint

   
    def __init__(self, start=None, end=None, unit=None, t=None, dt=None, tz=None):
        
        # Handle t and dt shortcuts
        if t:
            start=TimePoint(t=t, tz=tz)
        if dt:
            start=TimePoint(dt=dt, tz=tz)
        
        if not isinstance(start, self.__POINT_TYPE__):
            raise TypeError('Slot start must be a Point object (got "{}")'.format(start.__class__.__name__))

        try:
            if start.tz != end.tz:
                raise ValueError('{} start and end must have the same time zone (got start.tz="{}", end.tz="{}")'.format(self.__class__.__name__, start.tz, end.tz))
        except AttributeError:
            if end is None:
                # We are using the Unit, use the start
                self.tz = start.tz
            else:
                # If we don't have a time zone, we don't have TimePoints, the parent will make the Slot creation fail with a TypeError
                pass
        else:    
            self.tz = start.tz
        
        # Convert unit to TimeUnit
        if unit and not isinstance(unit, TimeUnit):
            unit = TimeUnit(unit)
                
        # Call parent init
        super(TimeSlot, self).__init__(start=start, end=end, unit=unit)
        
        # If we did not have the end, set its timezone now:
        if end is None:
            self.end.change_timezone(self.start.tz)


    # Overwrite parent succedes, this has better performance as it checks for only one dimension
    def __succedes__(self, other):
        if other.end.t != self.start.t:
            # Take into account floating point rounding errors
            if is_close(other.end.t, self.start.t):
                return True
            return False
        else:
            return True

    def change_timezone(self, new_timezone):
        self.start.change_timezone(new_timezone)
        self.end.change_timezone(new_timezone)
        self.tz = self.start.tz

    @property
    def unit(self):
        try:
            return self._unit
        except AttributeError:
            # Use the string representation to handle floating point seconds
            self._unit = TimeUnit(str(self.end.t-self.start.t)+'s')
            return self._unit
                   
    @property
    def t(self):
        return self.start.t
    
    @property
    def dt(self):
        return self.start.dt


class DataSlot(Slot):
    """A slot that carries some data."""

    def __init__(self, **kwargs):
        try:
            self._data = kwargs.pop('data')
        except KeyError:
            raise Exception('A DataSlot requires a special "data" argument (got only "{}")'.format(kwargs))

        try:
            data_loss = kwargs.pop('data_loss')
        except KeyError:
            pass
        else:
            self._data_loss = data_loss

        super(DataSlot, self).__init__(**kwargs)

    def __repr__(self):
        return '{} with start="{}" and end="{}"'.format(self.__class__.__name__, self.start, self.end)

    def __eq__(self, other):
        if self._data != other._data:
            return False
        return super(DataSlot, self).__eq__(other)

    @property
    def data(self):
        return self._data

    @property
    def data_loss(self):
        try:
            return self._data_loss
        except AttributeError:
            return None
    
    @property
    def data_reconstructed(self):
        try:
            return self._data_reconstructed
        except AttributeError:
            return None
    
    @data_reconstructed.setter
    def data_reconstructed(self, value):
        self._data_reconstructed = value


class DataTimeSlot(DataSlot, TimeSlot):
    """A slot that carries some data in the time dimension."""

    def __repr__(self):
        #if self.data_loss is not None:
        #    return '{} @ t=[{},{}] ([{},{}]) with data={} and data_loss={}'.format(self.__class__.__name__, self.start.t, self.end.t, self.start.dt, self.end.dt, self.data, self.data_loss)
        #else:
        #    return '{} @ t=[{},{}] ([{},{}]) with data={}'.format(self.__class__.__name__, self.start.t, self.end.t, self.start.dt, self.end.dt, self.data)

        if self.data_loss is not None:
            return '{} @ [{},{}] ([{},{}]) with data={} and data_loss={}'.format(self.__class__.__name__, self.start.t, self.end.t, self.start.dt, self.end.dt, self.data, self.data_loss)
        else:
            return '{} @ [{},{}] ([{},{}]) with data={}'.format(self.__class__.__name__, self.start.t, self.end.t, self.start.dt, self.end.dt, self.data)
        


#======================
#  Slot Series
#======================

class SlotSeries(Series):
    """A series of generic slots, where each item is guaranteed to be in succession."""
    
    __TYPE__ = Slot

    def append(self, item):
        
        # Slots can belong to the same series if they are in succession (tested with the __succedes__ method)
        # and if they have the same unit, which we test here instead as the __succedes__ is more general.
        try:
            if self._resolution != item.unit:
                # Try for floatign point precision errors
                abort = False
                try:
                    if not  is_close(self._resolution, item.unit):
                        abort = True
                except (TypeError, ValueError):
                    abort = True
                if abort:
                    raise ValueError('Cannot add items with different units (I have "{}" and you tried to add "{}")'.format(self._resolution, item.unit))
        except AttributeError:
            self._resolution = item.unit

        # Call parent append
        super(SlotSeries, self).append(item)



class TimeSlotSeries(SlotSeries):
    """A series of slots in time, where each item is guaranteed to be in succession."""
    
    __TYPE__ = TimeSlot

    def append(self, item):
        
        if not self.tz:
            # If no timezone set, use the item one's
            self._tz = item.tz
            
        else:
            # Else, check for the same time zone
            if self._tz != item.tz:
                raise ValueError('Cannot add items on different time zones (I have "{}" and you tried to add "{}")'.format(self.tz, item.start.tz))

        super(TimeSlotSeries, self).append(item)
        
    def change_timezone(self, new_timezone):
        for time_slot in self:
            time_slot.change_timezone(new_timezone)
        self._tz = time_slot.tz

    @property
    def resolution(self):
        return self._resolution

    @property
    def tz(self):
        try:
            return self._tz
        except AttributeError:
            return None



class DataSlotSeries(SlotSeries):
    """A series of data slots, where each item is guaranteed to be in succession and to carry the same data type."""

    __TYPE__ = DataSlot

    # Check data compatibility
    def append(self, item):
        
        # Check for data compatibility
        try:
            if not type(self.item_data_reference) == type(item.data):
                raise TypeError('Got different data: {} vs {}'.format(self.item_data_reference.__class__.__name__, item.data.__class__.__name__))
            if isinstance(self.item_data_reference, list):
                if len(self.item_data_reference) != len(item.data):
                    raise ValueError('Got different data lengths: {} vs {}'.format(len(self.item_data_reference), len(item.data)))
            if isinstance(self.item_data_reference, dict):
                if set(self.item_data_reference.keys()) != set(item.data.keys()):
                    raise ValueError('Got different data keys: {} vs {}'.format(self.item_data_reference.keys(), item.data.keys()))

        except AttributeError:
            # TODO: uniform self.tz, self._resolution, self.item_data_reference
            self.item_data_reference = item.data
        
        super(DataSlotSeries, self).append(item)
    
    def data_keys(self):
        if len(self) == 0:
            return None
        else:
            # TODO: can we optimize here? Computing them once and then serving them does not work if someone changes data keys...
            try:
                return sorted(list(self[0].data.keys()))
            except AttributeError:
                return list(range(len(self[0].data)))

    def remove_data_loss(self):
        new_series = self.duplicate()
        for item in new_series:
            item._data_loss = None
        return new_series


class DataTimeSlotSeries(DataSlotSeries, TimeSlotSeries):
    """A series of data slots in time, where each item is guaranteed to be in succession and to carry the same data type."""

    __TYPE__ = DataTimeSlot

    def __init__(self, *args, **kwargs):

        # Handle df kwarg
        df = kwargs.pop('df', None)
        
        # Handle first argument as dataframe
        if args and (isinstance(args[0], DataFrame)):
            df = args[0]

        # Do we have to initialize the time series from a dataframe?        
        if df is not None:

            # Create data time points list
            data_time_slots = []
            
            # Get data frame labels
            labels = list(df.columns) 

            # Get TimeUnit directly using and converting the inferred frequncy on the Pandas Data Frame
            unit_str=df.index.inferred_freq
            
            if not unit_str:
                raise Exception('Cannot infer the time unit for the slots')
            
            # Calendar
            unit_str=unit_str.replace('A', 'Y')    # Year (end) ?
            unit_str=unit_str.replace('Y', 'Y')    # Year (end) )
            unit_str=unit_str.replace('AS', 'Y')    # Year (start)
            unit_str=unit_str.replace('YS', 'Y')    # Year (start)
            unit_str=unit_str.replace('MS', 'M')    # Month (start)
            unit_str=unit_str.replace('M', 'M')    # Month (end) 
            unit_str=unit_str.replace('D', 'D')    # Day
            
            # Physical
            unit_str=unit_str.replace('H', 'h')    # Hour
            unit_str=unit_str.replace('T', 'm')    # Minute
            unit_str=unit_str.replace('min', 'm')  # Minute
            unit_str=unit_str.replace('S', 's')    # Second
            
            if len(unit_str) == 1:
                unit_str = '1'+unit_str
            logger.info('Assuming a slot time unit of "{}"'.format(unit_str))
            
            for row in df.iterrows():
                
                # Set the timestamp
                if not isinstance(row[0], datetime):
                    raise TypeError('A DataFrame with a DateTime index column is required')
                dt = row[0]
                
                # Prepare data
                data = {}
                for i,label in enumerate(labels):
                    data[label] = row[1][i]
                data_time_slots.append(DataTimeSlot(dt=dt, unit=TimeUnit(unit_str), data=data))
            
            # Set the list of data time points
            super(DataTimeSlotSeries, self).__init__(*data_time_slots, **kwargs)
            return None
        
        # Original init
        super(DataTimeSlotSeries, self).__init__(*args, **kwargs)

    @property
    def df(self):
        '''The time series as a Pandas DataFrame object'''
        data_keys = self.data_keys()
        
        if self[0].data_loss is not None:
            dump_data_loss = True
        else:
            dump_data_loss = False
        
        if dump_data_loss:
            columns = ['Timestamp'] + data_keys + ['data_loss']
        else:
            columns = ['Timestamp'] + data_keys
            
        df = DataFrame(columns=columns)
        for item in self:
            values = [item.data[key] for key in data_keys]
            if dump_data_loss:
                df = df.append(DataFrame([[item.dt]+values+[item.data_loss]], columns=columns))
            else:
                df = df.append(DataFrame([[item.dt]+values], columns=columns))                
        df = df.set_index('Timestamp')
        return df

    def plot(self, engine='dg', *args, **kwargs):
        if engine=='mp':
            from .plots import matplotlib_plot
            matplotlib_plot(self, *args, **kwargs)
        elif engine=='dg':
            from .plots import dygraphs_plot
            dygraphs_plot(self, *args, **kwargs)
        else:
            raise Exception('Unknown plotting engine "{}'.format(engine))

    def __repr__(self):
        if len(self):
            # TODO: "slots of unit" ?
            return 'Time series of #{} slots of {}, from slot starting @ {} ({}) to slot starting @ {} ({})'.format(len(self), self._resolution, self[0].start.t, self[0].start.dt, self[-1].start.t, self[-1].start.dt)            
        else:
            return 'Time series of #0 slots'
    
    
    